from __future__ import annotations

from enum import Enum
from typing import Callable, List, Optional, Union

import pandas as pd

from ..base import Composite, Pipelines, Transformations
from ..transformations.columns import SelectColumns
from ..transformations.dev import Identity
from ..utils.list import (
    flatten,
    has_intersection,
    keep_only_duplicates,
    wrap_in_double_list_if_needed,
    wrap_in_list,
)
from .common import get_concatenated_names


class ResolutionStrategy(Enum):
    """
    Parameters
    ----------
    first : str
        Only keep the first (leftmost) duplicate column(s).
    last : str
        Only keep the last (rightmost) duplicate column(s).
    both : str
        Keep both duplicate columns.
    """

    first = "first"
    last = "last"
    both = "both"

    @staticmethod
    def from_str(value: Union[str, ResolutionStrategy]) -> ResolutionStrategy:
        if isinstance(value, ResolutionStrategy):
            return value
        for strategy in ResolutionStrategy:
            if strategy.value == value:
                return strategy
        else:
            raise ValueError(f"Unknown ResolutionStrategy: {value}")


class Concat(Composite):
    """
    Concatenates the results of multiple pipelines.

    Parameters
    ----------

    pipelines : Pipelines
        A list of pipelines to be applied to the data, independently of each other.
    if_duplicate_keep : Union[ResolutionStrategy, str], optional
        How to handle duplicate columns, by default ResolutionStrategy.first

    Examples
    --------
        >>> from fold.loop import train_backtest
        >>> from fold.splitters import SlidingWindowSplitter
        >>> from fold.composites import Concat
        >>> from fold.utils.tests import generate_sine_wave_data
        >>> X, y  = generate_sine_wave_data()
        >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
        >>> pipeline = Concat([
        ...     lambda X: X.assign(sine_plus_1=X["sine"] + 1),
        ...     lambda X: X.assign(sine_plus_2=X["sine"] + 2),
        ... ])
        >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
        >>> preds.head()
                             sine_plus_1  sine_plus_2    sine
        2021-12-31 15:40:00       1.0000       2.0000 -0.0000
        2021-12-31 15:41:00       1.0126       2.0126  0.0126
        2021-12-31 15:42:00       1.0251       2.0251  0.0251
        2021-12-31 15:43:00       1.0377       2.0377  0.0377
        2021-12-31 15:44:00       1.0502       2.0502  0.0502
    """

    ResolutionStrategy = ResolutionStrategy
    properties = Composite.Properties()

    def __init__(
        self,
        pipelines: Pipelines,
        if_duplicate_keep: Union[ResolutionStrategy, str] = ResolutionStrategy.first,
    ) -> None:
        self.pipelines = pipelines
        self.if_duplicate_keep = ResolutionStrategy.from_str(if_duplicate_keep)
        self.name = "Concat-" + get_concatenated_names(pipelines)

    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        columns = flatten([result.columns.to_list() for result in results])
        duplicates = keep_only_duplicates(columns)
        if len(duplicates) == 0:
            return pd.concat(results, axis="columns")

        if len(duplicates) > 0 or self.if_duplicate_keep != ResolutionStrategy.both:
            duplicate_columns = [
                result[duplicates]
                for result in results
                if has_intersection(result.columns.to_list(), duplicates)
            ]
            results = [result.drop(columns=duplicates) for result in results]
            if self.if_duplicate_keep == ResolutionStrategy.first:
                return pd.concat(results + [duplicate_columns[0]], axis="columns")
            elif self.if_duplicate_keep == ResolutionStrategy.last:
                return pd.concat(results + [duplicate_columns[-1]], axis="columns")
            elif self.if_duplicate_keep == ResolutionStrategy.both:
                return pd.concat(results + duplicate_columns, axis="columns")
            else:
                raise ValueError(
                    f"ResolutionStrategy is not valid: {self.if_duplicate_keep}"
                )
        else:
            return pd.concat(results, axis="columns")

    def get_child_transformations_primary(self) -> Pipelines:
        return self.pipelines

    def clone(self, clone_child_transformations: Callable) -> Concat:
        return Concat(
            pipelines=clone_child_transformations(self.pipelines),
            if_duplicate_keep=self.if_duplicate_keep,
        )


class Pipeline(Composite):
    """
    An optional wrappers that is equivalent to using a single array for the transformations.
    It executes the transformations sequentially, in the order they are provided.

    Parameters
    ----------

    pipeline : Pipeline
        A list of transformations or models to be applied to the data.

    """

    properties = Composite.Properties(primary_only_single_pipeline=True)

    def __init__(
        self,
        pipeline: "Pipeline",
    ) -> None:
        self.pipeline = wrap_in_double_list_if_needed(pipeline)
        self.name = "Pipeline-" + get_concatenated_names(pipeline)

    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        return results[0]

    def get_child_transformations_primary(self) -> Pipelines:
        return self.pipeline

    def clone(self, clone_child_transformations: Callable) -> Pipeline:
        return Pipeline(pipeline=clone_child_transformations(self.pipeline))


def TransformColumn(
    columns: Union[List[str], str], pipeline: Transformations
) -> Composite:
    """
    Transforms a single or multiple columns using the given pipeline.
    """
    return Concat(
        [
            [SelectColumns(columns)] + wrap_in_list(pipeline),
            Identity(),
        ],
        if_duplicate_keep=ResolutionStrategy.first,
    )



from __future__ import annotations

from copy import deepcopy
from typing import List, Optional

import pandas as pd

from ..base import Composite, Transformation, Transformations
from ..models.base import Model
from ..utils.checks import is_prediction, is_X_available
from ..utils.trim import trim_initial_nans
from .backend import get_backend_dependent_functions
from .memory import postprocess_X_y_into_memory, preprocess_X_y_with_memory
from .types import Backend, Stage


def recursively_transform(
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
    transformations: Transformations,
    stage: Stage,
    backend: Backend,
) -> pd.DataFrame:
    """
    The main function to transform (and fit or update) a pipline of transformations.
    `stage` is used to determine whether to run the inner loop for online models.
    """
    if y is not None and len(X) != len(y):
        y = y[X.index]

    if isinstance(transformations, List):
        for transformation in transformations:
            X = recursively_transform(
                X, y, sample_weights, transformation, stage, backend
            )
        return X

    elif isinstance(transformations, Composite):
        return process_composite(transformations, X, y, sample_weights, stage, backend)

    elif isinstance(transformations, Transformation) or isinstance(
        transformations, Model
    ):
        # If the transformation needs to be "online", and we're in the update stage, we need to run the inner loop.
        if (
            transformations.properties.mode == Transformation.Properties.Mode.online
            and stage in [Stage.update, Stage.update_online_only]
            and not transformations.properties._internal_supports_minibatch_backtesting
        ):
            return process_with_inner_loop(transformations, X, y, sample_weights)
        # If the transformation is "online" but also supports our internal "mini-batch"-style updating
        elif (
            transformations.properties.mode == Transformation.Properties.Mode.online
            and stage in [Stage.update, Stage.update_online_only]
            and transformations.properties._internal_supports_minibatch_backtesting
        ):
            return process_internal_online_model_minibatch_inference_and_update(
                transformations, X, y, sample_weights
            )

        # or perform "mini-batch" updating OR the initial fit.
        else:
            return process_minibatch_transformation(
                transformations, X, y, sample_weights, stage
            )

    else:
        raise ValueError(
            f"{transformations} is not a Fold Transformation, but of type"
            f" {type(transformations)}"
        )


def process_composite(
    composite: Composite,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
    stage: Stage,
    backend: Backend,
) -> pd.DataFrame:
    backend_functions = get_backend_dependent_functions(backend)

    composite.before_fit(X)
    primary_transformations = composite.get_child_transformations_primary()

    results_primary = backend_functions.process_primary_child_transformations(
        __process_primary_child_transform,
        enumerate(primary_transformations),
        composite,
        X,
        y,
        sample_weights,
        stage,
        backend,
    )

    if composite.properties.primary_only_single_pipeline:
        assert len(results_primary) == 1, ValueError(
            "Expected single output from primary transformations, got"
            f" {len(results_primary)} instead."
        )
    if composite.properties.primary_requires_predictions:
        assert is_prediction(results_primary[0]), ValueError(
            "Expected predictions from primary transformations, but got something else."
        )

    secondary_transformations = composite.get_child_transformations_secondary()

    if secondary_transformations is None:
        return composite.postprocess_result_primary(results_primary, y)

    results_secondary = backend_functions.process_secondary_child_transformations(
        __process_secondary_child_transform,
        enumerate(secondary_transformations),
        composite,
        X,
        y,
        sample_weights,
        results_primary,
        stage,
        backend,
    )

    if composite.properties.secondary_only_single_pipeline:
        assert len(results_secondary) == 1, ValueError(
            "Expected single output from secondary transformations, got"
            f" {len(results_secondary)} instead."
        )
    if composite.properties.secondary_requires_predictions:
        assert is_prediction(results_secondary[0]), ValueError(
            "Expected predictions from secondary transformations, but got"
            " something else."
        )

    return composite.postprocess_result_secondary(results_primary, results_secondary, y)


def process_with_inner_loop(
    transformation: Transformation,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
) -> pd.DataFrame:
    if len(X) == 0:
        return pd.DataFrame()

    # We need to run the inference & fit loop on each row, sequentially (one-by-one).
    # This is so the transformation can update its parameters after each sample.

    def transform_row(
        X_row: pd.DataFrame, y_row: Optional[pd.Series], sample_weights_row
    ):
        X_row_with_memory, y_row_with_memory = preprocess_X_y_with_memory(
            transformation, X_row, y_row, in_sample=False
        )
        result = transformation.transform(X_row_with_memory, in_sample=False)
        if y_row is not None:
            transformation.update(
                X_row_with_memory, y_row_with_memory, sample_weights_row
            )
            postprocess_X_y_into_memory(
                transformation, X_row_with_memory, y_row_with_memory, False
            )
        return result.loc[X_row.index]

    return pd.concat(
        [
            transform_row(
                X.loc[index:index],
                y.loc[index:index] if y is not None else None,
                sample_weights.loc[index] if sample_weights is not None else None,
            )
            for index in X.index
        ],
        axis="index",
    )


def process_internal_online_model_minibatch_inference_and_update(
    transformation: Transformation,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
) -> pd.DataFrame:
    X, y = trim_initial_nans(X, y)
    X_with_memory, y_with_memory = preprocess_X_y_with_memory(
        transformation, X, y, in_sample=True
    )
    postprocess_X_y_into_memory(transformation, X_with_memory, y_with_memory, True)
    return_value = transformation.transform(X_with_memory, in_sample=True)

    transformation.update(X_with_memory, y_with_memory, sample_weights)
    postprocess_X_y_into_memory(transformation, X, y, False)
    return return_value.loc[X.index]


def process_minibatch_transformation(
    transformation: Transformation,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
    stage: Stage,
) -> pd.DataFrame:
    X, y = trim_initial_nans(X, y)

    if not is_X_available(X) and transformation.properties.requires_X:
        raise ValueError(
            "X is None, but transformation"
            f" {transformation.__class__.__name__} requires it."
        )

    in_sample = stage == Stage.inital_fit
    X_with_memory, y_with_memory = preprocess_X_y_with_memory(
        transformation, X, y, in_sample=in_sample
    )
    # The order is:
    # 1. fit (if we're in the initial_fit stage)
    if stage == Stage.inital_fit:
        transformation.fit(X_with_memory, y_with_memory, sample_weights)
        postprocess_X_y_into_memory(
            transformation,
            X_with_memory,
            y_with_memory,
            in_sample=stage == Stage.inital_fit,
        )
    # 2. transform (inference)
    X_with_memory, y_with_memory = preprocess_X_y_with_memory(
        transformation, X, y, in_sample=False
    )
    return_value = transformation.transform(X_with_memory, in_sample=in_sample)
    # 3. update (if we're in the update stage)
    if stage == Stage.update:
        transformation.update(X_with_memory, y_with_memory, sample_weights)
        postprocess_X_y_into_memory(transformation, X, y, False)
    return return_value.loc[X.index]


def __process_primary_child_transform(
    composite: Composite,
    index: int,
    child_transform: Transformations,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
    stage: Stage,
    backend: Backend,
) -> pd.DataFrame:
    X, y = composite.preprocess_primary(X, index, y, fit=stage.is_fit_or_update())
    return recursively_transform(X, y, sample_weights, child_transform, stage, backend)


def __process_secondary_child_transform(
    composite: Composite,
    index: int,
    child_transform: Transformations,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
    results_primary: List[pd.DataFrame],
    stage: Stage,
    backend: Backend,
) -> pd.DataFrame:
    X, y = composite.preprocess_secondary(
        X, y, results_primary, index, fit=stage.is_fit_or_update()
    )
    return recursively_transform(X, y, sample_weights, child_transform, stage, backend)


def deepcopy_pipelines(transformation: Transformations) -> Transformations:
    if isinstance(transformation, List):
        return [deepcopy_pipelines(t) for t in transformation]
    elif isinstance(transformation, Composite):
        return transformation.clone(deepcopy_pipelines)
    else:
        return deepcopy(transformation)



from typing import Optional, Tuple

import pandas as pd

from ..base import Transformation


def preprocess_X_y_with_memory(
    transformation: Transformation,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    in_sample: bool,
) -> Tuple[pd.DataFrame, pd.Series]:
    if transformation._state is None or transformation.properties.memory_size is None:
        return X, y
    memory_X, memory_y = transformation._state.memory_X, transformation._state.memory_y
    non_overlapping_indices = ~memory_X.index.isin(X.index)
    memory_X, memory_y = (
        memory_X[non_overlapping_indices],
        memory_y[non_overlapping_indices],
    )
    if len(memory_X) == 0:
        return X, y
    assert len(memory_X) == len(memory_y)
    if y is None:
        return (
            pd.concat(
                [memory_X.iloc[-transformation.properties.memory_size : None], X],
                axis="index",
            ),
            y,
        )
    elif in_sample is True or transformation.properties.memory_size == 0:
        memory_y.name = y.name
        return pd.concat([memory_X, X], axis="index"), pd.concat(
            [memory_y, y], axis="index"
        )
    else:
        memory_y.name = y.name
        return pd.concat(
            [memory_X.iloc[-transformation.properties.memory_size : None], X],
            axis="index",
        ), pd.concat(
            [memory_y.iloc[-transformation.properties.memory_size : None], y],
            axis="index",
        )


def postprocess_X_y_into_memory(
    transformation: Transformation,
    X: pd.DataFrame,
    y: pd.Series,
    in_sample: bool,
) -> None:
    # don't update the transformation if we're in inference mode (y is None)
    if transformation.properties.memory_size is None or y is None:
        return

    window_size = (
        len(X)
        if transformation.properties.memory_size == 0
        else transformation.properties.memory_size
    )
    if in_sample:
        # store the whole training X and y
        transformation._state = Transformation.State(
            memory_X=X,
            memory_y=y,
        )
    elif transformation.properties.memory_size < len(X):
        memory_X, memory_y = (
            transformation._state.memory_X,
            transformation._state.memory_y,
        )
        memory_y.name = y.name
        #  memory requirement is greater than the current batch, so we use the previous memory as well
        transformation._state = Transformation.State(
            memory_X=pd.concat([memory_X, X], axis="index").iloc[-window_size:None],
            memory_y=pd.concat([memory_y, y], axis="index").iloc[-window_size:None],
        )
    else:
        transformation._state = Transformation.State(
            memory_X=X.iloc[-window_size:None],
            memory_y=y.iloc[-window_size:None],
        )



from typing import List, Optional, Tuple, Union

import pandas as pd

from ..base import Composite, Pipeline, TrainedPipelines, Transformation
from ..splitters import Fold, SlidingWindowSplitter, Splitter
from ..utils.list import wrap_in_list
from .backend import get_backend_dependent_functions
from .checks import check_types
from .common import deepcopy_pipelines, recursively_transform
from .convenience import replace_transformation_if_not_fold_native
from .types import Backend, Stage, TrainMethod


def train(
    pipeline: Pipeline,
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splitter: Splitter,
    sample_weights: Optional[pd.Series] = None,
    train_method: Union[TrainMethod, str] = TrainMethod.parallel,
    backend: Union[Backend, str] = Backend.no,
    silent: bool = False,
) -> TrainedPipelines:
    """
    Trains a pipeline on a given dataset, for all folds returned by the Splitter.

    Parameters
    ----------
    pipeline: Pipeline
        The pipeline to be fitted.
    X: Optional[pd.DataFrame]
        Exogenous Data.
    y: pd.Series
        Endogenous Data (Target).
    splitter: Splitter
        Defines how the folds should be constructed.
    train_method : TrainMethod, str = TrainMethod.parallel
        The training methodology, by default `parallel`.
    backend: str, Backend = Backend.no
        The library/service to use for parallelization / distributed computing, by default `no`.
    sample_weights: Optional[pd.Series] = None
        Weights assigned to each sample/timestamp, that are passed into models that support it, by default None.
    silent: bool = False
        Wether the pipeline should print to the console, by default False.

    Returns
    -------
    TrainedPipelines
        The fitted pipelines, for all folds.
    """
    X, y = check_types(X, y)
    train_method = TrainMethod.from_str(train_method)
    backend = Backend.from_str(backend)

    if isinstance(splitter, SlidingWindowSplitter):
        assert train_method == TrainMethod.parallel, (
            "SlidingWindowSplitter is conceptually incompatible with"
            " TrainMethod.sequential"
        )

    pipeline = wrap_in_list(pipeline)
    pipeline = replace_transformation_if_not_fold_native(pipeline)

    splits = splitter.splits(length=len(y))
    backend_functions = get_backend_dependent_functions(backend)
    if len(splits) == 0:
        raise ValueError("No splits were generated by the Splitter.")

    if train_method == TrainMethod.parallel_with_search and len(splits) > 1:
        first_batch_index, first_batch_transformations = process_pipeline_window(
            X,
            y,
            sample_weights,
            pipeline,
            splits[0],
            never_update=True,
            backend=backend,
        )

        rest_idx, rest_transformations = zip(
            *backend_functions.train_transformations(
                process_pipeline_window,
                first_batch_transformations,
                X,
                y,
                sample_weights,
                splits[1:],
                False,
                backend,
                silent,
            )
        )
        processed_idx = [first_batch_index] + list(rest_idx)
        processed_pipelines = [first_batch_transformations] + list(rest_transformations)
    elif train_method == TrainMethod.parallel and len(splits) > 1:
        processed_idx, processed_pipelines = zip(
            *backend_functions.train_transformations(
                process_pipeline_window,
                pipeline,
                X,
                y,
                sample_weights,
                splits,
                True,
                backend,
                silent,
            )
        )

    else:
        processed_idx = []
        processed_pipelines = []
        processed_pipeline = pipeline
        for split in splits:
            processed_id, processed_pipeline = process_pipeline_window(
                X,
                y,
                sample_weights,
                processed_pipeline,
                split,
                False,
                backend,
            )
            processed_idx.append(processed_id)
            processed_pipelines.append(processed_pipeline)

    return [
        pd.Series(
            transformation_over_time,
            index=processed_idx,
            name=transformation_over_time[0].name,
        )
        for transformation_over_time in zip(*processed_pipelines)
    ]


def process_pipeline_window(
    X: pd.DataFrame,
    y: pd.Series,
    sample_weights: Optional[pd.Series],
    transformations: List[Union[Transformation, Composite]],
    split: Fold,
    never_update: bool,
    backend: Backend,
) -> Tuple[int, List[Union[Transformation, Composite]]]:
    stage = Stage.inital_fit if (split.order == 0 or never_update) else Stage.update
    window_start = (
        split.update_window_start if stage == Stage.update else split.train_window_start
    )
    window_end = (
        split.update_window_end if stage == Stage.update else split.train_window_end
    )
    X_train = X.iloc[window_start:window_end]
    y_train = y.iloc[window_start:window_end]

    sample_weights_train = (
        sample_weights.iloc[window_start:window_end]
        if sample_weights is not None
        else None
    )

    transformations = deepcopy_pipelines(transformations)
    X_train = recursively_transform(
        X_train, y_train, sample_weights_train, transformations, stage, backend
    )

    return split.model_index, transformations



from dataclasses import dataclass
from typing import List, Optional, Union


@dataclass
class Fold:
    order: int
    model_index: int
    train_window_start: int
    train_window_end: int
    update_window_start: int
    update_window_end: int
    test_window_start: int
    test_window_end: int


class Splitter:
    def splits(self, length: int) -> List[Fold]:
        raise NotImplementedError


def translate_float_if_needed(window_size: Union[int, float], length: int) -> int:
    if window_size >= 1 and isinstance(window_size, int):
        return window_size
    elif window_size < 1 and isinstance(window_size, float):
        return int(window_size * length)
    else:
        raise ValueError(
            "Invalid window size, should be either a float less than 1 or an int"
            " greater than 1"
        )


class SlidingWindowSplitter(Splitter):
    """
    Creates folds with a sliding window.
    The folds are created by moving the train and test windows forward by a fixed step size.
    See [the documentation](https://dream-faster.github.io/fold/concepts/splitters/) for more details.

    Parameters
    ----------

    initial_train_window : int, float
        The initial training window size. If a float, it is interpreted as a fraction of the total length of the data.
    step : int, float
        The step size of the sliding window. If a float, it is interpreted as a fraction of the total length of the data.
    embargo : int, optional
        The gap between the train and the test window, by default 0.
    start : int, optional
        The start index of the first fold, by default 0.
    end : int, optional
        The end index of the last fold, by default None.
    """

    def __init__(
        self,
        initial_train_window: Union[
            int, float
        ],  # this is what you don't out of sample get predictions for
        step: Union[int, float],
        embargo: int = 0,
        start: int = 0,
        end: Optional[int] = None,
    ) -> None:
        self.window_size = initial_train_window
        self.step = step
        self.embargo = embargo
        self.start = start
        self.end = end

    def splits(self, length: int) -> List[Fold]:
        end = self.end if self.end is not None else length
        window_size = translate_float_if_needed(self.window_size, length)
        step = translate_float_if_needed(self.step, length)
        return [
            Fold(
                order=order,
                model_index=index,
                train_window_start=index - window_size,
                train_window_end=index - self.embargo,
                update_window_start=0,  # SlidingWindowSplitter is incompatible with sequential updates
                update_window_end=0,
                test_window_start=index,
                test_window_end=min(end, index + step),
            )
            for order, index in enumerate(range(self.start + window_size, end, step))
        ]


class ExpandingWindowSplitter(Splitter):
    """
    Creates folds with an expanding window.
    The folds are created by moving the end of the train and test windows forward by a fixed step size,
    while keeping the training window's start fixed.
    See [the documentation](https://dream-faster.github.io/fold/concepts/splitters/) for more details.

    Parameters
    ----------

    initial_train_window : int, float
        The initial training window size. If a float, it is interpreted as a fraction of the total length of the data.
    step : int, float
        The step size of the sliding window. If a float, it is interpreted as a fraction of the total length of the data.
    embargo : int, optional
        The gap between the train and the test window, by default 0.
    start : int, optional
        The start index of the first fold, by default 0.
    end : int, optional
        The end index of the last fold, by default None.
    """

    def __init__(
        self,
        initial_train_window: Union[
            int, float
        ],  # this is what you don't out of sample get predictions for
        step: Union[int, float],
        embargo: int = 0,
        start: int = 0,
        end: Optional[int] = None,
    ) -> None:
        self.window_size = initial_train_window
        self.step = step
        self.embargo = embargo
        self.start = start
        self.end = end

    def splits(self, length: int) -> List[Fold]:
        end = self.end if self.end is not None else length
        window_size = translate_float_if_needed(self.window_size, length)
        step = translate_float_if_needed(self.step, length)

        return [
            Fold(
                order=order,
                model_index=index,
                train_window_start=self.start,
                train_window_end=index - self.embargo,
                update_window_start=index
                - step,  # the length of the update window is the step size, see documentation
                update_window_end=index - self.embargo,
                test_window_start=index,
                test_window_end=min(end, index + step),
            )
            for order, index in enumerate(range(self.start + window_size, end, step))
        ]


class SingleWindowSplitter(Splitter):
    """
    Creates a single fold with a fixed train and test window.
    See [the documentation](https://dream-faster.github.io/fold/concepts/splitters/) for more details.

    Parameters
    ----------
    train_window : int, float
        The training window size. If a float, it is interpreted as a fraction of the total length of the data.
    embargo : int, optional
        The gap between the train and the test window, by default 0.
    """

    def __init__(
        self,
        train_window: Union[
            int, float
        ],  # this is what you don't out of sample get predictions for
        embargo: int = 0,
    ) -> None:
        self.window_size = train_window
        self.embargo = embargo

    def splits(self, length: int) -> List[Fold]:
        window_size = translate_float_if_needed(self.window_size, length)

        return [
            Fold(
                order=0,
                model_index=0,
                train_window_start=0,
                train_window_end=window_size - self.embargo,
                update_window_start=0,
                update_window_end=0,  # SingleWindowSplitter is incompatible with sequantial updating
                test_window_start=window_size,
                test_window_end=length,
            ),
        ]



from __future__ import annotations

from typing import Any, Callable, List, Optional, Tuple

import pandas as pd

from fold.composites.common import get_concatenated_names
from fold.utils.list import wrap_in_double_list_if_needed

from ..base import Composite, Pipeline, Pipelines, T


class Sample(Composite):
    """
    Sample data with an imbalanced-learn sampler instance during training.
    No sampling is done during inference or backtesting.

    Warning:
    This seriously challenges the continuity of the data, which is very important for traditional time series models.
    Use with caution, and only with tabular ML models.

    Parameters
    ----------
    sampler : Any
        An imbalanced-learn sampler instance (subclass of `BaseSampler`).
    pipeline : Pipeline
        A pipeline to be applied to the sampled data.

    Examples
    --------
        >>> from fold.loop import train_backtest
        >>> from fold.splitters import SlidingWindowSplitter
        >>> from fold.composites import ModelResiduals
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from imblearn.under_sampling import RandomUnderSampler
        >>> from fold.utils.tests import generate_zeros_and_ones_skewed
        >>> X, y  = generate_zeros_and_ones_skewed()
        >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
        >>> pipeline = Sample(
        ...     sampler=RandomUnderSampler(),
        ...     pipeline=RandomForestClassifier(),
        ... )
        >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)


    References
    ----------
    [imbalanced-learn](https://imbalanced-learn.org/)
    """

    properties = Composite.Properties(
        primary_requires_predictions=False,
        primary_only_single_pipeline=True,
    )

    def __init__(
        self,
        sampler: Any,
        pipeline: Pipeline,
    ) -> None:
        self.sampler = sampler
        self.pipeline = wrap_in_double_list_if_needed(pipeline)
        self.name = f"Sample-{sampler.__class__.__name__}-{get_concatenated_names(self.pipeline)}"

    def preprocess_primary(
        self, X: pd.DataFrame, index: int, y: T, fit: bool
    ) -> Tuple[pd.DataFrame, T]:
        if fit:
            X_resampled, y_resampled = self.sampler.fit_resample(X, y)
            X_resampled.columns = X.columns
            X_resampled.index = X.index[: len(X_resampled)]
            if y is not None:
                y_resampled.name = y.name
                y_resampled.index = y.index[: len(y_resampled)]
            return X_resampled, y_resampled
        else:
            return X, y

    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        return results[0]

    def get_child_transformations_primary(self) -> Pipelines:
        return self.pipeline

    def clone(self, clone_child_transformations: Callable) -> Sample:
        return Sample(
            sampler=self.sampler,
            pipeline=clone_child_transformations(self.pipeline),
        )



from inspect import getfullargspec
from typing import Optional

import pandas as pd

from ..base import FeatureSelector, Transformation, fit_noop


class WrapSKLearnTransformation(Transformation):
    """
    Wraps an SKLearn Transformation.
    There's no need to use it directly, `fold` automatically wraps all sklearn transformations into this class.
    """

    properties = Transformation.Properties(requires_X=True)

    def __init__(self, transformation) -> None:
        if hasattr(transformation, "set_output"):
            transformation = transformation.set_output(transform="pandas")
        self.transformation = transformation
        self.name = transformation.__class__.__name__

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        fit_func = (
            self.transformation.partial_fit
            if hasattr(self.transformation, "partial_fit")
            else self.transformation.fit
        )

        argspec = getfullargspec(fit_func)
        if len(argspec.args) == 3:
            fit_func(X, y)
        elif len(argspec.args) == 4:
            fit_func(X, y, sample_weights)

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        if hasattr(self.transformation, "partial_fit"):
            argspec = getfullargspec(self.transformation.partial_fit)
            if len(argspec.args) == 3:
                self.transformation.partial_fit(X, y)
            elif len(argspec.args) == 4:
                self.transformation.partial_fit(X, y, sample_weights)
        # if we don't have partial_fit, we can't update the model (maybe throw an exception, and force user to wrap it into `DontUpdate`?)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if hasattr(self.transformation, "set_output"):
            return self.transformation.transform(X)
        else:
            return pd.DataFrame(self.transformation.transform(X), columns=X.columns)

    def inverse_transform(self, X: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(
            self.transformation.inverse_transform(X), columns=X.columns, index=X.index
        )


class WrapSKLearnFeatureSelector(FeatureSelector):
    """
    Wraps an SKLearn Feature Selector class, stores the selected columns in `selected_features` property.
    There's no need to use it directly, `fold` automatically wraps all sklearn feature selectors into this class.
    """

    properties = Transformation.Properties(requires_X=True)
    selected_features: Optional[str] = None

    def __init__(self, transformation) -> None:
        if hasattr(transformation, "set_output"):
            transformation = transformation.set_output(transform="pandas")
        self.transformation = transformation
        self.name = transformation.__class__.__name__

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        self.transformation.fit(X, y)
        if hasattr(self.transformation, "get_feature_names_out"):
            self.selected_features = self.transformation.get_feature_names_out()
        else:
            self.selected_features = X.columns[
                self.transformation.get_support()
            ].to_list()

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return X[self.selected_features]

    update = fit_noop



from __future__ import annotations

from typing import Union

import pandas as pd

from ..base import fit_noop
from .base import TimeSeriesModel


class Naive(TimeSeriesModel):
    """
    A univariate model that predicts the last target value.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.models import Naive
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
    >>> pipeline = Naive()
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> pd.concat([preds, y[preds.index]], axis=1).head()
                         predictions_Naive    sine
    2021-12-31 15:40:00            -0.0000  0.0126
    2021-12-31 15:41:00             0.0126  0.0251
    2021-12-31 15:42:00             0.0251  0.0377
    2021-12-31 15:43:00             0.0377  0.0502
    2021-12-31 15:44:00             0.0502  0.0628

    ```
    """

    name = "Naive"
    properties = TimeSeriesModel.Properties(
        requires_X=False,
        mode=TimeSeriesModel.Properties.Mode.online,
        memory_size=1,
        _internal_supports_minibatch_backtesting=False,
    )

    def predict(
        self, X: pd.DataFrame, past_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        # it's an online transformation, so len(X) will be always 1,
        return pd.Series(past_y.iloc[-1].squeeze(), index=X.index[-1:None])

    def predict_in_sample(
        self, X: pd.DataFrame, lagged_y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        return lagged_y

    fit = fit_noop
    update = fit



import importlib.util
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Tuple, Union

import pandas as pd
from sklearn.metrics import mean_squared_error

from ..base import OutOfSamplePredictions, Pipeline, TrainedPipelines
from ..splitters import Splitter
from .backtesting import backtest
from .training import train
from .types import Backend, TrainMethod

if TYPE_CHECKING:
    from krisi import ScoreCard


def backtest_score(
    trained_pipelines: TrainedPipelines,
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splitter: Splitter,
    backend: Union[str, Backend] = Backend.no,
    sample_weights: Optional[pd.Series] = None,
    silent: bool = False,
    krisi_args: Optional[Dict[str, Any]] = None,
    evaluation_func: Callable = mean_squared_error,
) -> Tuple[Union["ScoreCard", Dict[str, float]], OutOfSamplePredictions]:
    """
    Run backtest then scoring.
    If [`krisi`](https://github.com/dream-faster/krisi) is installed it will use it to generate a ScoreCard,
    otherwise it will run the `evaluation_func` passed in.

    Parameters
    ----------
    trained_pipelines: TrainedPipelines
        The fitted pipelines, for all folds.
    X: Optional[pd.DataFrame]
        Exogenous Data.
    y: pd.Series
        Endogenous Data (Target).
    splitter: Splitter
        Defines how the folds should be constructed.
    backend: str, Backend = Backend.no
        The library/service to use for parallelization / distributed computing, by default `no`.
    sample_weights: Optional[pd.Series] = None
        Weights assigned to each sample/timestamp, that are passed into models that support it, by default None.
    silent: bool = False
        Wether the pipeline should print to the console, by default False.
    krisi_args: Optional[Dict[str, Any]] = None
        Arguments that will be passed into `krisi` score function, by default None.
    evaluation_func: Callable = mean_squared_error
        Function to evaluate with if `krisi` is not available, by default `mean_squared_error`.


    Returns
    -------
    "ScoreCard", Dict[str, float]
        A ScoreCard if `krisi` is available, else the result of the `evaluation_func` in a dict
    OutOfSamplePredictions
        Predictions for all folds, concatenated.
    """

    pred = backtest(
        trained_pipelines,
        X,
        y,
        splitter,
        backend,
        sample_weights,
        silent,
        mutate=False,
    )

    if importlib.util.find_spec("krisi") is not None:
        from krisi import score

        scorecard = score(
            y[pred.index],
            pred.squeeze(),
            **(krisi_args if krisi_args is not None else {}),
        )
    else:
        scorecard = {
            "mean_squared_error": evaluation_func(y[pred.index], pred.squeeze())
        }
    return scorecard, pred


def train_backtest(
    pipeline: Pipeline,
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splitter: Splitter,
    backend: Union[Backend, str] = Backend.no,
    sample_weights: Optional[pd.Series] = None,
    train_method: Union[TrainMethod, str] = TrainMethod.parallel,
    silent: bool = False,
) -> Tuple[OutOfSamplePredictions, TrainedPipelines]:
    """
    Run train and backtest.

    Parameters
    ----------
    pipeline: Pipeline
        The pipeline to be fitted.
    X: Optional[pd.DataFrame]
        Exogenous Data.
    y: pd.Series
        Endogenous Data (Target).
    splitter: Splitter
        Defines how the folds should be constructed.
    backend: str, Backend = Backend.no
        The library/service to use for parallelization / distributed computing, by default `no`.
    sample_weights: Optional[pd.Series] = None
        Weights assigned to each sample/timestamp, that are passed into models that support it, by default None.
    train_method: TrainMethod, str = TrainMethod.parallel
        The training methodology, by default `parallel`.
    silent: bool = False
        Wether the pipeline should print to the console, by default False.

    Returns
    -------
    OutOfSamplePredictions
        Predictions for all folds, concatenated.
    TrainedPipelines
        The fitted pipelines, for all folds.
    """
    trained_pipelines = train(
        pipeline, X, y, splitter, sample_weights, train_method, backend, silent
    )

    pred = backtest(
        trained_pipelines,
        X,
        y,
        splitter,
        backend,
        sample_weights,
        silent,
    )

    return pred, trained_pipelines


def train_evaluate(
    pipeline: Pipeline,
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splitter: Splitter,
    backend: Backend = Backend.no,
    sample_weights: Optional[pd.Series] = None,
    train_method: Union[TrainMethod, str] = TrainMethod.parallel,
    silent: bool = False,
    krisi_args: Optional[Dict[str, Any]] = None,
    evaluation_func: Callable = mean_squared_error,
) -> Tuple[
    Union["ScoreCard", Dict[str, float]],
    OutOfSamplePredictions,
    TrainedPipelines,
]:
    """
    Run train, backtest then run scoring.
    If [`krisi`](https://github.com/dream-faster/krisi) is installed it will use it to generate a ScoreCard,
    otherwise it will run the `evaluation_func` passed in.

    Parameters
    ----------
    pipeline: Pipeline
        The pipeline to be fitted.
    X: pd.DataFrame, optional
        Exogenous Data.
    y: pd.Series
        Endogenous Data (Target).
    splitter: Splitter
        Defines how the folds should be constructed.
    backend: str, Backend = Backend.no
        The library/service to use for parallelization / distributed computing, by default `no`.
    sample_weights: pd.Series, optional = None
        Weights assigned to each sample/timestamp, that are passed into models that support it, by default None.
    train_method: TrainMethod, str = TrainMethod.parallel
        The training methodology, by default `parallel`.
    silent: bool = False
        Wether the pipeline should print to the console, by default False.
    krisi_args: Dict[str, Any], optional = None
        Arguments that will be passed into `krisi` score function, by default None.
    evaluation_func: Callable = mean_squared_error
        Function to evaluate with if `krisi` is not available, by default `mean_squared_error`.

    Returns
    -------
    "ScoreCard", Dict[str, float]
        A ScoreCard if `krisi` is available, else the result of the `evaluation_func` in a dict
    OutOfSamplePredictions
        Predictions for all folds, concatenated.
    TrainedPipelines
        The fitted pipelines, for all folds.
    """
    trained_pipelines = train(
        pipeline, X, y, splitter, sample_weights, train_method, backend, silent
    )

    scorecard, pred = backtest_score(
        trained_pipelines,
        X,
        y,
        splitter,
        backend,
        sample_weights,
        silent,
        krisi_args,
        evaluation_func,
    )

    return scorecard, pred, trained_pipelines



from random import choices
from typing import List, Tuple

import numpy as np
import pandas as pd

from fold.base import Transformation


def generate_sine_wave_data(
    cycles: int = 2, length: int = 1000, freq: str = "min"
) -> Tuple[pd.DataFrame, pd.Series]:
    end_value = np.pi * 2 * cycles
    my_wave = np.sin(np.linspace(0, end_value, length + 1))
    series = pd.Series(
        my_wave,
        name="sine",
        index=pd.date_range(end="2022", periods=len(my_wave), freq=freq),
    ).round(4)
    X = series.to_frame()
    y = series.shift(-1)[:-1]
    X = X[:-1]
    return X, y


def generate_all_zeros(length: int = 1000) -> Tuple[pd.DataFrame, pd.Series]:
    length += 1
    series = pd.Series(
        [0] * length,
        index=pd.date_range(end="2022", periods=length, freq="m"),
    )
    X = series.to_frame()
    y = series.shift(-1)[:-1]
    X = X[:-1]
    return X, y


def generate_zeros_and_ones(
    length: int = 1000, labels=[1, 0]
) -> Tuple[pd.DataFrame, pd.Series]:
    return generate_zeros_and_ones_skewed(length, labels, weights=[0.5, 0.5])


def generate_zeros_and_ones_skewed(
    length: int = 1000, labels=[1, 0], weights: List[float] = [0.2, 0.8]
) -> Tuple[pd.DataFrame, pd.Series]:
    length += 1
    series = pd.Series(
        choices(population=labels, weights=weights, k=length),
        index=pd.date_range(end="2022", periods=length, freq="s"),
    )
    X = series.to_frame()
    y = series.shift(-1)[:-1]
    X = X[:-1]
    return X, y


def generate_monotonous_data(
    length: int = 1000, freq: str = "m"
) -> Tuple[pd.DataFrame, pd.Series]:
    values = np.linspace(0, 1, num=length + 1)
    series = pd.Series(
        values,
        name="linear",
        index=pd.date_range(end="2022", periods=len(values), freq=freq),
    )
    X = series.to_frame()
    y = series.shift(-1)[:-1]
    X = X[:-1]
    return X, y


def check_if_transformation_mutates(
    X: pd.DataFrame, y: pd.Series, transformation: Transformation
) -> None:
    X_train = X[: len(X) // 2]
    y_train = y[: len(y) // 2]
    X_train_before = X_train.copy()
    y_train_before = y_train.copy()
    transformation.fit(X_train, y_train)
    assert X_train_before.equals(X_train)
    assert y_train_before.equals(y_train)

    X_test = X[len(X) // 2 :]
    y_test = y[len(y) // 2 :]
    X_test_before = X_test.copy()
    y_test_before = y_test.copy()
    transformation.transform(X_test, in_sample=True)
    transformation.update(X, y)
    assert X_test_before.equals(X_test)
    assert y_test_before.equals(y_test)

    X_val = X[-1:]
    y_val = y[-1:]
    X_val_before = X_val.copy()
    y_val_before = y_val.copy()
    transformation.transform(X_val, in_sample=False)
    assert X_val_before.equals(X_val)
    assert y_val_before.equals(y_val)



from importlib.util import find_spec
from typing import Callable, List

from sklearn.base import ClassifierMixin, RegressorMixin, TransformerMixin
from sklearn.feature_selection import SelectorMixin
from sklearn.pipeline import Pipeline as SKLearnPipeline

from fold.models.sklearn import (
    WrapSKLearnClassifier,
    WrapSKLearnPipeline,
    WrapSKLearnRegressor,
)
from fold.transformations.sklearn import (
    WrapSKLearnFeatureSelector,
    WrapSKLearnTransformation,
)

from ..base import Composite, Pipeline, Transformation
from ..transformations.function import WrapFunction


def replace_transformation_if_not_fold_native(
    transformation: Pipeline,
) -> Pipeline:
    if isinstance(transformation, List):
        return [replace_transformation_if_not_fold_native(t) for t in transformation]
    elif isinstance(transformation, RegressorMixin):
        return WrapSKLearnRegressor(transformation)
    elif isinstance(transformation, ClassifierMixin):
        return WrapSKLearnClassifier(transformation)
    elif isinstance(transformation, Callable):
        return WrapFunction(transformation, None)
    elif isinstance(transformation, SelectorMixin):
        return WrapSKLearnFeatureSelector(transformation)
    elif isinstance(transformation, TransformerMixin):
        return WrapSKLearnTransformation(transformation)
    elif isinstance(transformation, SKLearnPipeline):
        return WrapSKLearnPipeline(transformation)
    elif isinstance(transformation, Composite):
        return transformation.clone(replace_transformation_if_not_fold_native)
    elif isinstance(transformation, Transformation):
        return transformation
    elif find_spec("fold_wrappers") is not None:
        from fold_wrappers.convenience import wrap_transformation_if_possible

        return wrap_transformation_if_possible(transformation)
    else:
        raise ValueError(f"Transformation {transformation} is not supported")



from dataclasses import dataclass
from typing import Callable

from fold.loop.types import Backend


@dataclass
class BackendDependentFunctions:
    process_primary_child_transformations: Callable
    process_secondary_child_transformations: Callable
    train_transformations: Callable


def get_backend_dependent_functions(backend: Backend) -> BackendDependentFunctions:
    if backend == Backend.ray:
        from .ray import (
            process_primary_child_transformations,
            process_secondary_child_transformations,
            train_transformations,
        )

        return BackendDependentFunctions(
            process_primary_child_transformations,
            process_secondary_child_transformations,
            train_transformations,
        )
    elif backend == Backend.no:
        from .sequential import (
            process_primary_child_transformations,
            process_secondary_child_transformations,
            train_transformations,
        )

        return BackendDependentFunctions(
            process_primary_child_transformations,
            process_secondary_child_transformations,
            train_transformations,
        )



from __future__ import annotations

from copy import deepcopy
from typing import Callable, List, Optional, Tuple

import pandas as pd

from ..base import Composite, Pipeline, Pipelines, T
from ..utils.checks import all_have_probabilities
from ..utils.list import unique, wrap_in_double_list_if_needed
from .common import get_concatenated_names


class EnsembleEachColumn(Composite):
    """
    Train a pipeline for each column in the data, then ensemble their results.

    Parameters
    ----------
    pipeline: Pipeline
        Pipeline that get applied to every column, independently, their results then averaged.

    Examples
    --------
        >>> from fold.loop import train_backtest
        >>> from fold.splitters import SlidingWindowSplitter
        >>> from fold.composites import EnsembleEachColumn
        >>> from sklearn.ensemble import RandomForestRegressor
        >>> from fold.utils.tests import generate_sine_wave_data
        >>> X, y  = generate_sine_wave_data()
        >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
        >>> pipeline = EnsembleEachColumn(RandomForestRegressor())
        >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)

    """

    properties = Composite.Properties()
    pipelines_already_cloned = False

    def __init__(self, pipeline: Pipeline) -> None:
        self.pipelines: Pipelines = wrap_in_double_list_if_needed(pipeline)
        self.name = "PerColumnEnsemble-" + get_concatenated_names(self.pipelines)

    @classmethod
    def from_cloned_instance(
        cls, pipeline: Pipeline, pipelines_already_cloned: bool
    ) -> EnsembleEachColumn:
        instance = cls(pipeline=pipeline)
        instance.pipelines_already_cloned = pipelines_already_cloned
        return instance

    def before_fit(self, X: pd.DataFrame) -> None:
        if not self.pipelines_already_cloned:
            self.pipelines = [deepcopy(self.pipelines) for _ in X.columns]
            self.pipelines_already_cloned = True

    def preprocess_primary(
        self, X: pd.DataFrame, index: int, y: T, fit: bool
    ) -> Tuple[pd.DataFrame, pd.Series]:
        X = X.iloc[:, index].to_frame()
        return X, y

    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        return postprocess_results(results, self.name)

    def get_child_transformations_primary(self) -> Pipelines:
        return self.pipelines

    def clone(self, clone_child_transformations: Callable) -> EnsembleEachColumn:
        return EnsembleEachColumn.from_cloned_instance(
            pipeline=clone_child_transformations(self.pipelines),
            pipelines_already_cloned=self.pipelines_already_cloned,
        )


class TransformEachColumn(Composite):
    """
    Apply a single pipeline to each column, separately.

    Parameters
    ----------
    pipeline: Pipeline
        Pipeline that gets applied to each column

    Examples
    --------
        >>> from fold.loop import train_backtest
        >>> from fold.splitters import SlidingWindowSplitter
        >>> from fold.composites import TransformEachColumn
        >>> from sklearn.ensemble import RandomForestRegressor
        >>> from fold.utils.tests import generate_sine_wave_data
        >>> X, y  = generate_sine_wave_data()
        >>> X["sine_plus_1"] = X["sine"] + 1.0
        >>> X.head()
                               sine  sine_plus_1
        2021-12-31 07:20:00  0.0000       1.0000
        2021-12-31 07:21:00  0.0126       1.0126
        2021-12-31 07:22:00  0.0251       1.0251
        2021-12-31 07:23:00  0.0377       1.0377
        2021-12-31 07:24:00  0.0502       1.0502
        >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
        >>> pipeline = TransformEachColumn(lambda x: x + 1.0)
        >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
        >>> preds.head()
                               sine  sine_plus_1
        2021-12-31 15:40:00  1.0000       2.0000
        2021-12-31 15:41:00  1.0126       2.0126
        2021-12-31 15:42:00  1.0251       2.0251
        2021-12-31 15:43:00  1.0377       2.0377
        2021-12-31 15:44:00  1.0502       2.0502
    """

    properties = Composite.Properties()
    pipeline_already_cloned = False

    def __init__(self, pipeline: Pipeline) -> None:
        self.pipeline = wrap_in_double_list_if_needed(pipeline)
        self.name = "PerColumnTransform-" + get_concatenated_names(self.pipeline)

    @classmethod
    def from_cloned_instance(
        cls, pipeline: Pipeline, pipeline_already_cloned: bool
    ) -> TransformEachColumn:
        instance = cls(pipeline=pipeline)
        instance.pipeline_already_cloned = pipeline_already_cloned
        return instance

    def before_fit(self, X: pd.DataFrame) -> None:
        if not self.pipeline_already_cloned:
            self.pipeline = [deepcopy(self.pipeline) for _ in X.columns]
            self.pipeline_already_cloned = True

    def preprocess_primary(
        self, X: pd.DataFrame, index: int, y: T, fit: bool
    ) -> Tuple[pd.DataFrame, T]:
        return X.iloc[:, index].to_frame(), y

    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        return pd.concat(results, axis="columns")

    def get_child_transformations_primary(self) -> Pipelines:
        return self.pipeline

    def clone(self, clone_child_transformations: Callable) -> TransformEachColumn:
        return TransformEachColumn.from_cloned_instance(
            pipeline=clone_child_transformations(self.pipeline),
            pipeline_already_cloned=self.pipeline_already_cloned,
        )


class SkipNA(Composite):
    """
    Skips rows with NaN values in the input data.
    In the output, rows with NaNs are returned as is, all other rows transformed.

    Warning:
    This seriously challenges the continuity of the data, which is very important for traditional time series models.
    Use with caution, and only with tabular ML models.

    Parameters
    ----------
    pipeline: Pipeline
        Pipeline to run without NA values.

    Examples
    --------
        >>> from fold.loop import train_backtest
        >>> import numpy as np
        >>> from fold.splitters import SlidingWindowSplitter
        >>> from fold.composites import ModelResiduals
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from imblearn.under_sampling import RandomUnderSampler
        >>> from fold.utils.tests import generate_zeros_and_ones
        >>> X, y  = generate_zeros_and_ones()
        >>> X[1:100] = np.nan
        >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
        >>> pipeline = SkipNA(
        ...     pipeline=RandomForestClassifier(),
        ... )
        >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)

    """

    properties = Composite.Properties()

    def __init__(self, pipeline: Pipeline) -> None:
        self.pipeline = wrap_in_double_list_if_needed(pipeline)
        self.name = "SkipNA-" + get_concatenated_names(self.pipeline)

    def preprocess_primary(
        self, X: pd.DataFrame, index: int, y: T, fit: bool
    ) -> Tuple[pd.DataFrame, T]:
        self.original_index = X.index.copy()
        self.isna = X.isna().any(axis=1)
        return X[~self.isna], y[~self.isna] if y is not None else None

    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        results = [result.reindex(self.original_index) for result in results]
        return pd.concat(results, axis="columns")

    def get_child_transformations_primary(self) -> Pipelines:
        return self.pipeline

    def clone(self, clone_child_transformations: Callable) -> SkipNA:
        return SkipNA(
            pipeline=clone_child_transformations(self.pipeline),
        )


def postprocess_results(
    results: List[pd.DataFrame],
    name: str,
) -> pd.DataFrame:
    if all_have_probabilities(results):
        return get_groupped_columns_classification(results, name)
    else:
        return get_groupped_columns_regression(results, name)


def get_groupped_columns_regression(
    results: List[pd.DataFrame],
    name: str,
) -> pd.DataFrame:
    return (
        pd.concat(
            [
                df[
                    [col for col in df.columns if col.startswith("predictions_")]
                ].squeeze()
                for df in results
            ],
            axis="columns",
        )
        .mean(axis="columns")
        .rename(f"predictions_{name}")
        .to_frame()
    )


def get_groupped_columns_classification(
    results: List[pd.DataFrame],
    name: str,
) -> pd.DataFrame:
    columns = results[0].columns.to_list()
    probabilities_columns = [col for col in columns if col.startswith("probabilities_")]
    classes = unique([line.split("_")[-1] for line in probabilities_columns])

    predictions = (
        pd.concat(
            [
                df[
                    [col for col in df.columns if col.startswith("predictions_")]
                ].squeeze()
                for df in results
            ],
            axis="columns",
        )
        .mean(axis="columns")
        .rename(f"predictions_{name}")
    )

    probabilities = [
        (
            pd.concat(
                [
                    df[
                        [
                            col
                            for col in df.columns
                            if col.startswith("probabilities_")
                            and col.split("_")[-1] == selected_class
                        ]
                    ].squeeze()
                    for df in results
                ],
                axis="columns",
            )
            .mean(axis="columns")
            .rename(f"probabilities_{name}_{selected_class}")
        )
        for selected_class in classes
    ]
    return pd.concat([predictions] + probabilities, axis="columns")



from __future__ import annotations

from typing import Callable, List, Optional, Tuple, Union

import pandas as pd

from fold.composites.common import get_concatenated_names
from fold.utils.checks import get_prediction_column
from fold.utils.list import wrap_in_double_list_if_needed

from ..base import Composite, Pipeline, Pipelines, T


class MetaLabeling(Composite):
    """
    MetaLabeling takes a primary pipeline and a meta pipeline.
    The primary pipeline is used to predict the target variable.
    The meta pipeline is used to predict whether the primary model's prediction's are correct (a binary classification problem).
    It multiplies the probabilities from the meta pipeline with the predictions of the primary pipeline.

    It's only applicable for binary classification problems, where the labels are either `1`, `-1` or one of them are zero.

    Parameters
    ----------

    primary : Pipeline
        A pipeline to be applied to the data. Target (`y`) is unchanged.
    meta : Pipeline
        A pipeline to be applied to predict whether the primary pipeline's predictions are correct. Target (`y`) is `preds == y`.
    positive_class : int, float
        The positive class of the primary pipeline.
    primary_output_included :  bool, optional
        Whether the primary pipeline's output is included in the meta pipeline's input, by default False.


    Examples
    --------
        >>> from fold.loop import train_backtest
        >>> from fold.splitters import SingleWindowSplitter
        >>> from fold.composites import MetaLabeling
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from sklearn.linear_model import LogisticRegression
        >>> from fold.utils.tests import generate_zeros_and_ones
        >>> X, y  = generate_zeros_and_ones()
        >>> splitter = SingleWindowSplitter(train_window=0.5)
        >>> pipeline = MetaLabeling(
        ...     primary=LogisticRegression(),
        ...     meta=RandomForestClassifier(),
        ...     positive_class=1.0,
        ... )
        >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)


    Outputs
    -------
        A prediction is a float between -1 or 0, and 1.
        It does not output probabilities, as the prediction already includes that information.


    References
    ----------
    [Meta Labeling (A Toy Example)](https://hudsonthames.org/meta-labeling-a-toy-example/)
    [Meta-Labeling: Theory and Framework](https://jfds.pm-research.com/content/4/3/31)

    """

    properties = Composite.Properties(
        primary_requires_predictions=True,
        primary_only_single_pipeline=True,
        secondary_requires_predictions=True,
        secondary_only_single_pipeline=True,
    )

    def __init__(
        self,
        primary: Pipeline,
        meta: Pipeline,
        positive_class: Union[int, float],
        primary_output_included: bool = False,
    ) -> None:
        self.primary = wrap_in_double_list_if_needed(primary)
        self.meta = wrap_in_double_list_if_needed(meta)
        self.positive_class = (
            int(positive_class) if isinstance(positive_class, float) else positive_class
        )
        self.primary_output_included = primary_output_included
        self.name = "MetaLabeling-" + get_concatenated_names(self.primary + self.meta)

    def preprocess_secondary(
        self,
        X: pd.DataFrame,
        y: T,
        results_primary: List[pd.DataFrame],
        index: int,
        fit: bool,
    ) -> Tuple[pd.DataFrame, T]:
        X = (
            pd.concat([X] + results_primary, axis="columns")
            if self.primary_output_included
            else X
        )
        predictions = get_prediction_column(results_primary[0])
        y = y.astype(int) == predictions.astype(int)
        return X, y

    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        raise NotImplementedError

    def postprocess_result_secondary(
        self,
        primary_results: List[pd.DataFrame],
        secondary_results: List[pd.DataFrame],
        y: Optional[pd.Series],
    ) -> pd.DataFrame:
        primary_predictions = get_prediction_column(primary_results[0])
        meta_probabilities = secondary_results[0][
            [
                col
                for col in secondary_results[0].columns
                if col.startswith("probabilities_")
            ]
        ]
        meta_probabilities_positive_class = meta_probabilities[
            [
                col
                for col in meta_probabilities.columns
                if get_int_class(col.split("_")[-1]) == self.positive_class
            ]
        ]
        if len(meta_probabilities_positive_class.columns) != 1:
            raise ValueError(
                "Meta pipeline needs to be concluded with probabilities of the"
                f" positive class: {str(self.positive_class)}"
            )
        result = (
            primary_predictions * meta_probabilities_positive_class.squeeze()
        ).rename(f"predictions_{self.name}")
        dc = {
            col: f"probabilities_{self.name}_" + col.split("_")[-1]
            for col in meta_probabilities.columns
        }
        meta_probabilities = meta_probabilities.rename(columns=dc)
        return pd.concat([result, meta_probabilities], axis="columns")

    def get_child_transformations_primary(self) -> Pipelines:
        return self.primary

    def get_child_transformations_secondary(
        self,
    ) -> Optional[Pipelines]:
        return self.meta

    def clone(self, clone_child_transformations: Callable) -> MetaLabeling:
        return MetaLabeling(
            primary=clone_child_transformations(self.primary),
            meta=clone_child_transformations(self.meta),
            positive_class=self.positive_class,
            primary_output_included=self.primary_output_included,
        )


def get_int_class(input: str) -> int:
    if input.endswith(".0"):
        return int(float(input[:-2]))
    elif input == "True":
        return 1
    elif input == "False":
        return 0
    else:
        return int(input)



from typing import List

from ..base import Composite, Pipelines, Transformations
from ..utils.list import flatten, wrap_in_list


def get_flat_list_of_transformations(
    transformations: Pipelines,
) -> List[Transformations]:
    def get_all_transformations(transformations: Pipelines) -> Transformations:
        if isinstance(transformations, List):
            return [get_all_transformations(t) for t in transformations]
        elif isinstance(transformations, Composite):
            secondary_transformations = (
                []
                if (transformations.get_child_transformations_secondary()) is None
                else transformations.get_child_transformations_secondary()
            )
            return [
                get_all_transformations(child_t)
                for child_t in transformations.get_child_transformations_primary()
            ] + [
                get_all_transformations(child_t)
                for child_t in secondary_transformations
            ]
        else:
            return transformations

    return flatten(wrap_in_list(get_all_transformations(transformations)))


def get_concatenated_names(transformations: Pipelines) -> str:
    return "-".join(
        [
            transformation.name if hasattr(transformation, "name") else ""
            for transformation in get_flat_list_of_transformations(transformations)
        ]
    )



from ..base import Composite
from .columns import EnsembleEachColumn, SkipNA, TransformEachColumn
from .concat import Concat, Pipeline, TransformColumn
from .ensemble import Ensemble
from .metalabeling import MetaLabeling
from .residual import ModelResiduals
from .sample import Sample
from .select import SelectBest
from .target import TransformTarget



from typing import Dict, Optional, Union

import numpy as np
import pandas as pd

from fold.base import InvertibleTransformation, fit_noop


class TakeLog(InvertibleTransformation):
    """
    Takes the logarithm of the data.

    Parameters
    ----------
    base : int, str, optional
        The base of the logarithm, by default "e".
        Valid values are "e", np.e, "10", 10, "2", 2.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import TakeLog
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data(freq="min")
    >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
    >>> pipeline = TakeLog()
    >>> X["sine"].head()
    2021-12-31 07:20:00    0.0000
    2021-12-31 07:21:00    0.0126
    2021-12-31 07:22:00    0.0251
    2021-12-31 07:23:00    0.0377
    2021-12-31 07:24:00    0.0502
    Freq: T, Name: sine, dtype: float64
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds["sine"].head()
    2021-12-31 15:40:00        -inf
    2021-12-31 15:41:00   -4.374058
    2021-12-31 15:42:00   -3.684887
    2021-12-31 15:43:00   -3.278095
    2021-12-31 15:44:00   -2.991740
    Freq: T, Name: sine, dtype: float64

    ```
    """

    name = "TakeLog"
    properties = InvertibleTransformation.Properties(requires_X=True)

    def __init__(
        self,
        base: Union[int, str] = "e",
    ) -> None:
        if base not in ["e", np.e, "10", 10, "2", 2]:
            raise ValueError("base should be either 'e', np.e, '10', 10, '2', 2.")
        self.base = base

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if self.base == "e" or self.base == np.e:
            return pd.DataFrame(np.log(X.values), columns=X.columns, index=X.index)
        elif self.base == "10" or self.base == 10:
            return pd.DataFrame(np.log10(X.values), columns=X.columns, index=X.index)
        elif self.base == "2" or self.base == 2:
            return pd.DataFrame(np.log2(X.values), columns=X.columns, index=X.index)
        else:
            raise ValueError(f"Invalid base: {self.base}")

    def inverse_transform(self, X: pd.Series) -> pd.Series:
        if self.base == "e" or self.base == np.e:
            return pd.Series(np.exp(X.values), index=X.index)
        elif self.base == "10" or self.base == 10:
            return 10**X
        elif self.base == "2" or self.base == 2:
            return 2**X
        else:
            raise ValueError(f"Invalid base: {self.base}")

    fit = fit_noop
    update = fit_noop


class AddConstant(InvertibleTransformation):

    """
    Adds a constant to the data.

    Parameters
    ----------
    constant: int, float, Dict[str, Union[float, int]]
        The constant to add to the data. If a dictionary is passed, the values will be added to the columns with the same name.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddConstant
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data(freq="min")
    >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
    >>> pipeline = AddConstant(1.0)
    >>> X["sine"].head()
    2021-12-31 07:20:00    0.0000
    2021-12-31 07:21:00    0.0126
    2021-12-31 07:22:00    0.0251
    2021-12-31 07:23:00    0.0377
    2021-12-31 07:24:00    0.0502
    Freq: T, Name: sine, dtype: float64
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds["sine"].head()
    2021-12-31 15:40:00    1.0000
    2021-12-31 15:41:00    1.0126
    2021-12-31 15:42:00    1.0251
    2021-12-31 15:43:00    1.0377
    2021-12-31 15:44:00    1.0502
    Freq: T, Name: sine, dtype: float64

    ```
    """

    name = "AddConstant"
    properties = InvertibleTransformation.Properties(requires_X=True)

    def __init__(
        self,
        constant: Union[int, float, Dict[str, Union[float, int]]],
    ) -> None:
        if not isinstance(constant, (int, float, dict)):
            raise ValueError(
                "constant can be only integer, float or a dictionary of integers or"
                " floats"
            )

        self.constant = constant

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if isinstance(self.constant, dict):
            transformed_columns = X[list(self.constant.keys())] + pd.Series(
                self.constant
            )
            return pd.concat(
                [X.drop(columns=self.constant.keys()), transformed_columns],
                axis="columns",
            )
        else:
            return X + self.constant

    def inverse_transform(self, X: pd.Series) -> pd.Series:
        constant = self.constant
        if constant is dict:
            constant = next(iter(constant.values()))
        return X - constant

    fit = fit_noop
    update = fit_noop


class TurnPositive(InvertibleTransformation):
    """
    Adds a constant to the data, varying by column, so that all values are positive.
    It identifies the constant during training, and applies it during inference (and backtesting).
    Therefore there's no guarantee that the data will be positive during inference (and backtesting).

    It can not be updated after the initial training, as that'd change the underlying distribution of the data.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import TurnPositive
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data(freq="min")
    >>> X, y  = X - 1, y - 1
    >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
    >>> pipeline = TurnPositive()
    >>> X["sine"].head()
    2021-12-31 07:20:00   -1.0000
    2021-12-31 07:21:00   -0.9874
    2021-12-31 07:22:00   -0.9749
    2021-12-31 07:23:00   -0.9623
    2021-12-31 07:24:00   -0.9498
    Freq: T, Name: sine, dtype: float64
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds["sine"].head()
    2021-12-31 15:40:00    2.0000
    2021-12-31 15:41:00    2.0126
    2021-12-31 15:42:00    2.0251
    2021-12-31 15:43:00    2.0377
    2021-12-31 15:44:00    2.0502
    Freq: T, Name: sine, dtype: float64

    ```
    """

    name = "TurnPositive"
    properties = InvertibleTransformation.Properties(requires_X=True)

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        min_values = X.min(axis=0)
        self.constant = dict(min_values[min_values <= 0].abs() + 1)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        transformed_columns = X[list(self.constant.keys())] + pd.Series(self.constant)
        return pd.concat(
            [X.drop(columns=self.constant.keys()), transformed_columns],
            axis="columns",
        )

    def inverse_transform(self, X: pd.Series) -> pd.Series:
        return X - next(iter(self.constant.values()))

    update = fit_noop



from __future__ import annotations

from enum import Enum
from typing import Callable, List, Tuple, Union

import pandas as pd

from ..base import Transformation, fit_noop
from ..utils.list import wrap_in_list


class PredefinedFunction(Enum):
    mean = "mean"
    sum = "sum"
    median = "median"
    std = "std"
    var = "var"
    kurt = "kurt"
    min = "min"
    max = "max"
    corr = "corr"
    cov = "cov"
    skew = "skew"
    sem = "sem"

    @staticmethod
    def from_str(value: Union[str, PredefinedFunction]) -> PredefinedFunction:
        if isinstance(value, PredefinedFunction):
            return value
        for strategy in PredefinedFunction:
            if strategy.value == value:
                return strategy
        else:
            raise ValueError(f"Unknown PredefinedFunction: {value}")


ColumnOrColumns = Union[str, List[str]]
FunctionOrPredefined = Union[Callable, PredefinedFunction, str]
ColumnWindowFunction = Tuple[ColumnOrColumns, int, FunctionOrPredefined]


class AddWindowFeatures(Transformation):

    """
    Creates rolling window features on the specified columns.
    Equivalent to adding a new column by running: `df[column].rolling(window).function()`.


    Parameters
    ----------
    column_window_func : ColumnWindowFunction, List[ColumnWindowFunction]
        A list of tuples, where each tuple contains the column name, the window size and the function to apply.
        The function can be a predefined function (see PredefinedFunction) or a Callable (with a single parameter).

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddWindowFeatures
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
    >>> pipeline = AddWindowFeatures(("sine", 10, "mean"))
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                           sine  sine_10_mean
    2021-12-31 15:40:00 -0.0000      -0.05649
    2021-12-31 15:41:00  0.0126      -0.04394
    2021-12-31 15:42:00  0.0251      -0.03139
    2021-12-31 15:43:00  0.0377      -0.01883
    2021-12-31 15:44:00  0.0502      -0.00628

    ```
    """

    name = "AddWindowFeatures"

    def __init__(
        self,
        column_window_func: Union[ColumnWindowFunction, List[ColumnWindowFunction]],
    ) -> None:
        self.column_window_func = [
            (
                wrap_in_list(column),
                window,
                function
                if isinstance(function, Callable)
                else PredefinedFunction.from_str(function),
            )
            for column, window, function in wrap_in_list(column_window_func)
        ]
        max_memory = max([window for _, window, _ in self.column_window_func])
        self.properties = Transformation.Properties(
            requires_X=True, memory_size=max_memory
        )

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        X_function_applied = pd.DataFrame([], index=X.index)
        for columns, window, function in self.column_window_func:
            if isinstance(function, PredefinedFunction):
                function = getattr(pd.core.window.rolling.Rolling, function.value)
            function_name = (
                function.__name__ if function.__name__ != "<lambda>" else "transformed"
            )
            if columns[0] == "all":
                X_function_applied = pd.concat(
                    [
                        X_function_applied,
                        function(X.rolling(window)).add_suffix(
                            f"_{window}_{function_name}"
                        ),
                    ],
                    axis="columns",
                )
            else:
                for col in columns:
                    X_function_applied[f"{col}_{window}_{function_name}"] = function(
                        X[col].rolling(window)
                    )
        return pd.concat([X, X_function_applied], axis="columns")

    fit = fit_noop
    update = fit



from __future__ import annotations

import enum
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable, List, Optional, Tuple, TypeVar, Union

import pandas as pd

T = TypeVar("T", Optional[pd.Series], pd.Series)


class Composite(ABC):
    """
    A Composite contains other transformations.
    """

    @dataclass
    class Properties:
        primary_requires_predictions: bool = (
            False  # Primary pipeline need output from a model
        )
        primary_only_single_pipeline: bool = (
            False  # There should be a single primary pipeline.
        )
        secondary_requires_predictions: bool = (
            False  # Secondary pipeline need output from a model
        )
        secondary_only_single_pipeline: bool = (
            False  # There should be a single secondary pipeline.
        )

    properties: Properties

    @abstractmethod
    def get_child_transformations_primary(self) -> Pipelines:
        raise NotImplementedError

    def get_child_transformations_secondary(
        self,
    ) -> Optional[Pipelines]:
        return None

    @abstractmethod
    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        raise NotImplementedError

    def postprocess_result_secondary(
        self,
        primary_results: List[pd.DataFrame],
        secondary_results: List[pd.DataFrame],
        y: Optional[pd.Series],
    ) -> pd.DataFrame:
        raise NotImplementedError

    @abstractmethod
    def clone(self, clone_child_transformations: Callable) -> "Composite":
        raise NotImplementedError

    def before_fit(self, X: pd.DataFrame) -> None:
        pass

    def preprocess_primary(
        self, X: pd.DataFrame, index: int, y: T, fit: bool
    ) -> Tuple[pd.DataFrame, T]:
        return X, y

    def preprocess_secondary(
        self,
        X: pd.DataFrame,
        y: T,
        results_primary: List[pd.DataFrame],
        index: int,
        fit: bool,
    ) -> Tuple[pd.DataFrame, T]:
        return X, y


class Transformation(ABC):
    """
    A transformation is a single step in a pipeline.
    """

    @dataclass
    class Properties:
        class ModelType(enum.Enum):
            regressor = "regressor"
            classifier = "classifier"

        class Mode(enum.Enum):
            minibatch = "minibatch"
            online = "online"

        requires_X: bool
        mode: Mode = Mode.minibatch
        memory_size: Optional[
            int
        ] = None  # if not `None`, will inject past window with size of `memory` to update() & transformation(). if `0`, it'll remember all data. during the in_sample period, it'll contain all data.
        model_type: Optional[ModelType] = None
        _internal_supports_minibatch_backtesting: bool = False  # internal, during backtesting, calls predict_in_sample() instead of predict()

    @dataclass
    class State:
        memory_X: pd.DataFrame
        memory_y: pd.Series

    properties: Properties
    name: str
    _state: Optional[State] = None

    @abstractmethod
    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        """
        Called once, with on initial training window.
        """
        raise NotImplementedError

    @abstractmethod
    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        """
        Subsequent calls to update the model.
        """
        raise NotImplementedError

    @abstractmethod
    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        raise NotImplementedError


class InvertibleTransformation(Transformation, ABC):
    @abstractmethod
    def inverse_transform(self, X: pd.Series) -> pd.Series:
        raise NotImplementedError


class FeatureSelector(Transformation):
    selected_features: List[str]


Transformations = Union[
    Transformation,
    "Composite",
    List[Union[Transformation, "Composite"]],
]

Pipeline = Union[
    Union[Transformation, "Composite"], List[Union[Transformation, "Composite"]]
]
"""A list of `fold` objects that are executed sequentially. Or a single object."""
Pipelines = List[Pipeline]
"""Multiple, independent `Pipeline`s."""

TrainedPipelines = List[pd.Series]
"""A list of trained `Pipeline`s, to be used for backtesting."""
OutOfSamplePredictions = pd.DataFrame
"""The backtest's resulting out-of-sample predictions."""


def fit_noop(
    self,
    X: pd.DataFrame,
    y: pd.Series,
    sample_weights: Optional[pd.Series] = None,
) -> None:
    pass


class SingleFunctionTransformation(Transformation):
    properties = Transformation.Properties(requires_X=True)

    def get_function(self) -> Callable:
        raise NotImplementedError

    fit = fit_noop
    update = fit_noop

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return pd.concat([X, self.get_function()(X)], axis="columns")



from .base import Model
from .baseline import Naive
from .dummy import DummyClassifier, DummyRegressor
from .random import RandomClassifier
from .sklearn import WrapSKLearnClassifier, WrapSKLearnRegressor



from random import choices
from typing import List, Union

import numpy as np
import pandas as pd

from ..base import Transformation, fit_noop
from .base import Model


class RandomClassifier(Model):
    """
    A model that predicts random classes and probabilities.

    Parameters
    ----------
    all_classes : List[int]
        All possible classes.
    probability_mean : float
        The mean of the normal distribution used to generate the probabilities.


    Examples
    --------
    ```pycon
    >>> import numpy as np
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.models import RandomClassifier
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
    >>> np.random.seed(42)
    >>> pipeline = RandomClassifier([0,1], 0.5)
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)

    ```
    """

    properties = Transformation.Properties(requires_X=False)
    name = "RandomClassifier"

    def __init__(self, all_classes: List[int], probability_mean: float = 0.5) -> None:
        self.all_classes = all_classes
        self.probability_mean = probability_mean

    fit = fit_noop

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        predictions = pd.Series(
            choices(population=self.all_classes, k=len(X)),
            index=X.index,
            name="predictions_RandomClassifier",
        )
        probabilities = [
            pd.Series(
                np.random.normal(self.probability_mean, 0.1, len(X)).clip(0, 1),
                index=X.index,
                name=f"probabilities_RandomClassifier_{associated_class}",
            )
            for associated_class in self.all_classes
        ]

        return pd.concat([predictions] + probabilities, axis="columns")

    predict_in_sample = predict

    update = fit



from ..base import Transformation
from ..composites.concat import TransformColumn
from .columns import (
    DropColumns,
    OnlyPredictions,
    OnlyProbabilities,
    RenameColumns,
    SelectColumns,
)
from .date import (
    AddDateTimeFeatures,
    AddDayOfWeek,
    AddDayOfYear,
    AddHour,
    AddMinute,
    AddMonth,
    AddQuarter,
    AddSecond,
    AddWeek,
    AddWeekOfYear,
    AddYear,
)
from .difference import Difference
from .function import WrapFunction
from .holidays import AddHolidayFeatures
from .lags import AddLagsX, AddLagsY
from .math import AddConstant, TakeLog, TurnPositive
from .sklearn import WrapSKLearnFeatureSelector, WrapSKLearnTransformation
from .update import DontUpdate
from .window import AddWindowFeatures



from typing import List, Tuple, Union

import pandas as pd

from fold.utils.checks import is_X_available

from ..base import Transformation, fit_noop
from ..utils.list import flatten, transform_range_to_list, wrap_in_list


class AddLagsY(Transformation):
    """
    Adds past values of `y`.

    Parameters
    ----------
    lags : List[int], range
        A list of lags (of the target) to add as features.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddLagsY
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> _, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
    >>> pipeline = AddLagsY([1,2,3])
    >>> preds, trained_pipeline = train_backtest(pipeline, None, y, splitter)
    >>> preds.head()
                         y_lag_1  y_lag_2  y_lag_3
    2021-12-31 15:40:00  -0.0000  -0.0126  -0.0251
    2021-12-31 15:41:00   0.0126  -0.0000  -0.0126
    2021-12-31 15:42:00   0.0251   0.0126  -0.0000
    2021-12-31 15:43:00   0.0377   0.0251   0.0126
    2021-12-31 15:44:00   0.0502   0.0377   0.0251

    ```
    """

    def __init__(self, lags: Union[List[int], range]) -> None:
        if not isinstance(lags, range) and not isinstance(lags, List):
            raise ValueError("lags must be a range or a List")
        self.lags = sorted(transform_range_to_list(lags))
        self.name = f"AddLagsY-{self.lags}"
        self.properties = Transformation.Properties(
            requires_X=False,
            mode=Transformation.Properties.Mode.online,
            memory_size=max(self.lags),
            _internal_supports_minibatch_backtesting=True,
        )

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        lags = pd.DataFrame([], index=X.index)

        if in_sample:
            for lag in self.lags:
                lags[f"y_lag_{lag}"] = self._state.memory_y.shift(lag)[-len(X) :]
        else:
            past_y = self._state.memory_y.reindex(X.index)
            for lag in self.lags:
                lags[f"y_lag_{lag}"] = past_y.shift(lag)[-len(X) :]

        if is_X_available(X):
            return pd.concat([X, lags], axis="columns")
        else:
            # If X is just an DataFrame with zeros, then just return the lags
            return lags

    fit = fit_noop
    update = fit_noop


class AddLagsX(Transformation):
    """
    Adds past values of `X` for the desired column(s).

    Parameters
    ----------
    columns_and_lags : List[ColumnAndLag], ColumnAndLag
        A tuple (or a list of tuples) of the column name and a single or a list of lags to add as features.

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import AddLagsX
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data()
    >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
    >>> pipeline = AddLagsX([("sine", 1), ("sine", [2,3])])
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds.head()
                           sine  sine_lag_1  sine_lag_2  sine_lag_3
    2021-12-31 15:40:00 -0.0000     -0.0126     -0.0251     -0.0377
    2021-12-31 15:41:00  0.0126     -0.0000     -0.0126     -0.0251
    2021-12-31 15:42:00  0.0251      0.0126     -0.0000     -0.0126
    2021-12-31 15:43:00  0.0377      0.0251      0.0126     -0.0000
    2021-12-31 15:44:00  0.0502      0.0377      0.0251      0.0126

    ```
    """

    ColumnAndLag = Tuple[str, Union[int, List[int]]]

    def __init__(
        self, columns_and_lags: Union[List[ColumnAndLag], ColumnAndLag]
    ) -> None:
        self.columns_and_lags = wrap_in_list(columns_and_lags)

        def check_and_transform_if_needed(
            column_and_lag: AddLagsX.ColumnAndLag,
        ) -> AddLagsX.ColumnAndLag:
            column, lags = column_and_lag
            if (
                not isinstance(lags, int)
                and not isinstance(lags, List)
                and not isinstance(lags, range)
            ):
                raise ValueError("lags must be an int or a List or a range")
            lags = sorted(
                transform_range_to_list([lags] if isinstance(lags, int) else lags)
            )
            return column, lags

        self.columns_and_lags = list(
            map(check_and_transform_if_needed, self.columns_and_lags)
        )
        self.name = f"AddLagsX-{self.columns_and_lags}"
        self.properties = Transformation.Properties(
            requires_X=True,
            memory_size=max(flatten([l for _, l in self.columns_and_lags])),
        )

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        X_lagged = pd.DataFrame([], index=X.index)
        for column, lags in self.columns_and_lags:
            for lag in lags:
                if column == "all":
                    X_lagged = pd.concat(
                        [X_lagged, X.shift(lag)[-len(X) :].add_suffix(f"_lag_{lag}")],
                        axis="columns",
                    )
                else:
                    X_lagged[f"{column}_lag_{lag}"] = X[column].shift(lag)[-len(X) :]
        return pd.concat([X, X_lagged], axis="columns")

    fit = fit_noop
    update = fit_noop



from __future__ import annotations

from typing import Callable, List, Optional

import pandas as pd

from ..base import Composite, Pipelines
from .columns import postprocess_results
from .common import get_concatenated_names


class Ensemble(Composite):
    """
    Ensemble (average) the results of multiple pipelines.

    Parameters
    ----------

    pipelines : Pipelines
        A list of pipelines to be applied to the data, independently of each other.

    Examples
    --------
        >>> from fold.loop import train_backtest
        >>> from fold.splitters import SlidingWindowSplitter
        >>> from fold.composites import Ensemble
        >>> from fold.models import DummyRegressor
        >>> from fold.utils.tests import generate_sine_wave_data
        >>> X, y  = generate_sine_wave_data()
        >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
        >>> pipeline = Ensemble([
        ...     DummyRegressor(0.1),
        ...     DummyRegressor(0.9),
        ... ])
        >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
        >>> preds.squeeze().head()
        2021-12-31 15:40:00    0.5
        2021-12-31 15:41:00    0.5
        2021-12-31 15:42:00    0.5
        2021-12-31 15:43:00    0.5
        2021-12-31 15:44:00    0.5
        Freq: T, Name: predictions_Ensemble-DummyRegressor-DummyRegressor, dtype: float64
    """

    properties = Composite.Properties()

    def __init__(self, pipelines: Pipelines) -> None:
        self.pipelines = pipelines
        self.name = "Ensemble-" + get_concatenated_names(pipelines)

    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        return postprocess_results(results, self.name)

    def get_child_transformations_primary(self) -> Pipelines:
        return self.pipelines

    def clone(self, clone_child_transformations: Callable) -> Ensemble:
        return Ensemble(
            pipelines=clone_child_transformations(self.pipelines),
        )



from abc import abstractmethod
from typing import Union

import numpy as np
import pandas as pd

from ..base import Transformation


class Model(Transformation):
    @abstractmethod
    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        """
        Predictions for exclusively out-of-sample data, the model has never seen before.
        """
        raise NotImplementedError

    @abstractmethod
    def predict_in_sample(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        """
        Predictions for in-sample, already seen data.
        """
        raise NotImplementedError

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if in_sample:
            return postpostprocess_prediction(self.predict_in_sample(X), self.name)
        else:
            return postpostprocess_prediction(self.predict(X), self.name)


class TimeSeriesModel(Transformation):
    """
    Convenience class, for models that have access to past `y` values.
    """

    @abstractmethod
    def predict(
        self,
        X: pd.DataFrame,
        past_y: pd.Series,
    ) -> Union[pd.Series, pd.DataFrame]:
        """
        Predictions for exclusively out-of-sample data, the model has never seen before.
        """
        raise NotImplementedError

    @abstractmethod
    def predict_in_sample(
        self, X: pd.DataFrame, y: pd.Series
    ) -> Union[pd.Series, pd.DataFrame]:
        """
        Predictions for in-sample, already seen data.
        """
        raise NotImplementedError

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if in_sample:
            return postpostprocess_prediction(
                self.predict_in_sample(X, self._state.memory_y.shift(1)), self.name
            )
        else:
            return postpostprocess_prediction(
                self.predict(
                    X[-self.properties.memory_size + 1 : None],
                    pd.Series(
                        np.concatenate(
                            [
                                np.ones((1,)) * np.nan,
                                self._state.memory_y[
                                    -self.properties.memory_size : None
                                ],
                            ]
                        ),
                        index=X.index,
                    ),
                ),
                self.name,
            )


def postpostprocess_prediction(
    dataframe_or_series: Union[pd.DataFrame, pd.Series], name: str
) -> pd.DataFrame:
    if isinstance(dataframe_or_series, pd.Series):
        return dataframe_or_series.rename(f"predictions_{name}").to_frame()
    elif isinstance(dataframe_or_series, pd.DataFrame):
        if len(dataframe_or_series.columns) == 1:
            dataframe_or_series.columns = [f"predictions_{name}"]
        return dataframe_or_series
    else:
        raise ValueError(
            f"Expected dataframe or series, got {type(dataframe_or_series)} instead."
        )



from typing import Optional, Union

import pandas as pd
from tqdm.auto import tqdm

from ..base import OutOfSamplePredictions, TrainedPipelines
from ..splitters import Fold, Splitter
from ..utils.trim import trim_initial_nans_single
from .checks import check_types
from .common import deepcopy_pipelines, recursively_transform
from .types import Backend, Stage


def backtest(
    trained_pipelines: TrainedPipelines,
    X: Optional[pd.DataFrame],
    y: pd.Series,
    splitter: Splitter,
    backend: Union[Backend, str] = Backend.no,
    sample_weights: Optional[pd.Series] = None,
    silent: bool = False,
    mutate: bool = False,
) -> OutOfSamplePredictions:
    """
    Run backtest on TrainedPipelines and given data.

    Parameters
    ----------

    trained_pipelines: TrainedPipelines
        The fitted pipelines, for all folds.
    X: pd.DataFrame, optional
        Exogenous Data.
    y: pd.Series
        Endogenous Data (Target).
    splitter: Splitter
        Defines how the folds should be constructed.
    backend: str, Backend = Backend.no
        The library/service to use for parallelization / distributed computing, by default `no`.
    sample_weights: pd.Series, optional = None
        Weights assigned to each sample/timestamp, that are passed into models that support it, by default None.
    silent: bool = False
        Wether the pipeline should print to the console, by default False.
    mutate: bool = False
        Whether `trained_pipelines` should be mutated, by default False. This is discouraged.

    Returns
    -------
    OutOfSamplePredictions
        Predictions for all folds, concatenated.
    """
    backend = Backend.from_str(backend)
    X, y = check_types(X, y)

    results = [
        __backtest_on_window(
            trained_pipelines,
            split,
            X,
            y,
            sample_weights,
            backend,
            mutate=mutate,
        )
        for split in tqdm(splitter.splits(length=len(X)), disable=silent)
    ]
    return trim_initial_nans_single(pd.concat(results, axis="index"))


def __backtest_on_window(
    trained_pipelines: TrainedPipelines,
    split: Fold,
    X: pd.DataFrame,
    y: pd.Series,
    sample_weights: Optional[pd.Series],
    backend: Backend,
    mutate: bool,
) -> pd.DataFrame:
    current_pipeline = [
        pipeline_over_time.loc[split.model_index]
        for pipeline_over_time in trained_pipelines
    ]
    if not mutate:
        current_pipeline = deepcopy_pipelines(current_pipeline)

    X_test = X.iloc[split.test_window_start : split.test_window_end]
    y_test = y.iloc[split.test_window_start : split.test_window_end]
    sample_weights_test = (
        sample_weights.iloc[split.train_window_start : split.test_window_end]
        if sample_weights is not None
        else None
    )
    return recursively_transform(
        X_test,
        y_test,
        sample_weights_test,
        current_pipeline,
        stage=Stage.update_online_only,
        backend=backend,
    )



from typing import Optional

import pandas as pd

from ..base import Transformation, fit_noop


class DontUpdate(Transformation):
    properties = Transformation.Properties(requires_X=False)

    def __init__(self, transformation: Transformation) -> None:
        self.transformation = transformation
        self.name = f"DontUpdate-{transformation.name}"

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        self.transformation.fit(X, y, sample_weights)

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return self.transformation.transform(X, in_sample)

    update = fit_noop



from __future__ import annotations

from typing import Callable, List, Optional, Tuple

import pandas as pd

from fold.composites.common import get_concatenated_names
from fold.utils.checks import get_prediction_column
from fold.utils.list import wrap_in_double_list_if_needed

from ..base import Composite, Pipeline, Pipelines, T


class ModelResiduals(Composite):
    """
    This is a composite that combines two pipelines:
    * The primary pipeline is used to predict the target variable.
    * The meta pipeline is used to predict the primary pipeline's residual (or, error).

    It adds together the primary pipeline's output with the predicted residual.

    Also known as:
    - Residual chasing
    - Residual boosting
    - Hybrid approach
    - "Moving Average" in ARIMA

    It's only applicable for regression tasks.

    Parameters
    ----------

    primary : Pipeline
        A pipeline to be applied to the data. The target (`y`) is unchanged.
    meta : Pipeline
        A pipeline to predict the primary pipeline's residual. The target (`y`) is the primary pipeline's residual (or, error).
    primary_output_included : bool, optional
        Whether the primary pipeline's output is included in the meta pipeline's input, by default False.

    Examples
    --------
        >>> from fold.loop import train_backtest
        >>> from fold.splitters import SlidingWindowSplitter
        >>> from fold.composites import ModelResiduals
        >>> from sklearn.ensemble import RandomForestRegressor
        >>> from sklearn.linear_model import LinearRegression
        >>> from fold.utils.tests import generate_sine_wave_data
        >>> X, y  = generate_sine_wave_data()
        >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
        >>> pipeline = ModelResiduals(
        ...     primary=LinearRegression(),
        ...     meta=RandomForestRegressor(),
        ... )
        >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)

    References
    ----------
    - https://www.kaggle.com/code/ryanholbrook/hybrid-models
    - https://www.uber.com/en-DE/blog/m4-forecasting-competition/
    """

    properties = Composite.Properties(
        primary_requires_predictions=True,
        primary_only_single_pipeline=True,
        secondary_requires_predictions=True,
        secondary_only_single_pipeline=True,
    )

    def __init__(
        self,
        primary: Pipeline,
        meta: Pipeline,
        primary_output_included: bool = False,
    ) -> None:
        self.primary = wrap_in_double_list_if_needed(primary)
        self.meta = wrap_in_double_list_if_needed(meta)
        self.primary_output_included = primary_output_included
        self.name = "Hybrid-" + get_concatenated_names(self.primary + self.meta)

    def preprocess_primary(
        self, X: pd.DataFrame, index: int, y: T, fit: bool
    ) -> Tuple[pd.DataFrame, T]:
        return X, y

    def preprocess_secondary(
        self,
        X: pd.DataFrame,
        y: T,
        results_primary: List[pd.DataFrame],
        index: int,
        fit: bool,
    ) -> Tuple[pd.DataFrame, T]:
        X = (
            pd.concat([X] + results_primary, axis="columns")
            if self.primary_output_included
            else X
        )
        predictions = get_prediction_column(results_primary[0])
        residuals = y - predictions
        return X, residuals

    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        raise NotImplementedError

    def postprocess_result_secondary(
        self,
        primary_results: List[pd.DataFrame],
        secondary_results: List[pd.DataFrame],
        y: Optional[pd.Series],
    ) -> pd.DataFrame:
        primary_predictions = get_prediction_column(primary_results[0])
        residual_predictions = get_prediction_column(secondary_results[0])

        return (
            (primary_predictions + residual_predictions)
            .rename(f"predictions_{self.name}")
            .to_frame()
        )

    def get_child_transformations_primary(self) -> Pipelines:
        return self.primary

    def get_child_transformations_secondary(
        self,
    ) -> Optional[Pipelines]:
        return self.meta

    def clone(self, clone_child_transformations: Callable) -> ModelResiduals:
        return ModelResiduals(
            primary=clone_child_transformations(self.primary),
            meta=clone_child_transformations(self.meta),
            primary_output_included=self.primary_output_included,
        )



from typing import Callable, List, Optional

import pandas as pd
from tqdm.auto import tqdm

from ...base import Composite, Transformations
from ...splitters import Fold
from ..types import Backend, Stage


def train_transformations(
    func: Callable,
    transformations: Transformations,
    X: pd.DataFrame,
    y: pd.Series,
    sample_weights: Optional[pd.Series],
    splits: List[Fold],
    never_update: bool,
    backend: Backend,
    silent: bool,
):
    return [
        func(X, y, sample_weights, transformations, split, never_update, backend)
        for split in tqdm(splits, disable=silent)
    ]


def process_primary_child_transformations(
    func: Callable,
    list_of_child_transformations_with_index: List,
    composite: Composite,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
    stage: Stage,
    backend: Backend,
):
    return [
        func(
            composite, index, child_transformation, X, y, sample_weights, stage, backend
        )
        for index, child_transformation in list_of_child_transformations_with_index
    ]


def process_secondary_child_transformations(
    func: Callable,
    list_of_child_transformations_with_index: List,
    composite: Composite,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
    results_primary: List[pd.DataFrame],
    stage: Stage,
    backend: Backend,
):
    return [
        func(
            composite,
            index,
            child_transformation,
            X,
            y,
            sample_weights,
            results_primary,
            stage,
            backend,
        )
        for index, child_transformation in list_of_child_transformations_with_index
    ]



from typing import Tuple, Union

import numpy as np
import pandas as pd


def trim_initial_nans(X: pd.DataFrame, y: pd.Series) -> Tuple[pd.DataFrame, pd.Series]:
    # Optimize for speed, if the first value is not NaN, we can save all the subsequent computation
    if not X.iloc[0].isna().any() and (y is None or not np.isnan(y.iloc[0])):
        return X, y
    first_valid_index_X = get_first_valid_index(X)
    first_valid_index_y = get_first_valid_index(y)
    if first_valid_index_X is None or first_valid_index_y is None:
        return pd.DataFrame(), pd.Series()
    first_valid_index = max(first_valid_index_X, first_valid_index_y)
    return X.iloc[first_valid_index:], y.iloc[first_valid_index:]


def trim_initial_nans_single(X: pd.DataFrame) -> pd.DataFrame:
    # Optimize for speed, if the first value is not NaN, we can save all the subsequent computation
    if not X.iloc[0].isna().any():
        return X
    first_valid_index = get_first_valid_index(X)
    return X.iloc[first_valid_index:]


def get_first_valid_index(series: Union[pd.Series, pd.DataFrame]) -> int:
    if series.empty:
        return 0
    if isinstance(series, pd.DataFrame):
        return next(
            (
                idx
                for idx, (_, x) in enumerate(series.iterrows())
                if not pd.isna(x).any()
            ),
            None,
        )
    elif isinstance(series, pd.Series):
        return next(
            (idx for idx, (_, x) in enumerate(series.items()) if not pd.isna(x)),
            None,
        )



from __future__ import annotations

from enum import Enum
from typing import Optional, Tuple, Union

import pandas as pd


def load_dataset(
    dataset_name: str,
    base_path: str = "https://raw.githubusercontent.com/dream-faster/datasets/main/datasets",
) -> pd.DataFrame:
    return pd.read_csv(f"{base_path}/{dataset_name}.csv", parse_dates=True, index_col=0)


class DeduplicationStrategy(Enum):
    first = "first"
    last = "last"

    @staticmethod
    def from_str(value: Union[str, DeduplicationStrategy]) -> DeduplicationStrategy:
        if isinstance(value, DeduplicationStrategy):
            return value
        for strategy in DeduplicationStrategy:
            if strategy.value == value:
                return strategy
        else:
            raise ValueError(f"Unknown DeduplicationStrategy: {value}")


def __process_dataset(
    df: pd.DataFrame,
    target_col: str,
    resample: Optional[str] = None,
    deduplication_strategy: Optional[Union[DeduplicationStrategy, str]] = None,
    shorten: Optional[int] = None,
) -> Tuple[pd.DataFrame, pd.Series]:
    if deduplication_strategy is not None:
        df = df[
            ~df.index.duplicated(
                keep=DeduplicationStrategy.from_str(deduplication_strategy).value
            )
        ]
    if resample is not None:
        df = df.resample(resample).last()
    if shorten is not None:
        df = df[: shorten + 1]
    y = df[target_col].shift(-1)[:-1]
    X = df[:-1]

    return X, y


def get_preprocessed_dataset(
    dataset_name: str,
    target_col: str,
    base_path: str = "https://raw.githubusercontent.com/dream-faster/datasets/main/datasets",
    resample: Optional[str] = None,
    deduplication_strategy: Optional[Union[DeduplicationStrategy, str]] = None,
    shorten: Optional[int] = None,
) -> Tuple[pd.DataFrame, pd.Series]:
    return __process_dataset(
        load_dataset(dataset_name, base_path),
        target_col=target_col,
        resample=resample,
        deduplication_strategy=deduplication_strategy,
        shorten=shorten,
    )



from typing import Callable, Optional

import pandas as pd

from ..base import Transformation, fit_noop


class WrapFunction(Transformation):
    """
    Wraps and arbitrary function that will run at inference.
    """

    def __init__(self, func: Callable, past_window_size: Optional[int]) -> None:
        self.func = func
        self.name = func.__name__
        self.properties = Transformation.Properties(
            requires_X=True, memory_size=past_window_size
        )

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return self.func(X)

    fit = fit_noop
    update = fit



from typing import Optional

import pandas as pd

from ..base import InvertibleTransformation


class Difference(InvertibleTransformation):
    """
    Performs differencing.
    Sesonal differencing can be achieved by setting `lag` to the seasonality of the data.
    To achieve second-order differencing, simply chain multiple `Difference` transformations.

    Parameters
    ----------
    lag : int, optional
        the seasonality of the data, by default 1

    Examples
    --------
    ```pycon
    >>> from fold.loop import train_backtest
    >>> from fold.splitters import SlidingWindowSplitter
    >>> from fold.transformations import Difference
    >>> from fold.utils.tests import generate_sine_wave_data
    >>> X, y  = generate_sine_wave_data(freq="min")
    >>> splitter = SlidingWindowSplitter(initial_train_window=0.5, step=0.2)
    >>> pipeline = Difference()
    >>> X["sine"].head()
    2021-12-31 07:20:00    0.0000
    2021-12-31 07:21:00    0.0126
    2021-12-31 07:22:00    0.0251
    2021-12-31 07:23:00    0.0377
    2021-12-31 07:24:00    0.0502
    Freq: T, Name: sine, dtype: float64
    >>> preds, trained_pipeline = train_backtest(pipeline, X, y, splitter)
    >>> preds["sine"].head()
    2021-12-31 15:40:00    0.0126
    2021-12-31 15:41:00    0.0126
    2021-12-31 15:42:00    0.0125
    2021-12-31 15:43:00    0.0126
    2021-12-31 15:44:00    0.0125
    Freq: T, Name: sine, dtype: float64

    ```

    References
    ----------

    [Stationarity and differencing](https://otexts.com/fpp2/stationarity.html)
    """

    properties = InvertibleTransformation.Properties(requires_X=False)
    name = "Difference"

    def __init__(self, lag: int = 1) -> None:
        self.lag = lag

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        self.last_rows_X = X.iloc[-self.lag : None]

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        if len(X) >= self.lag:
            self.last_rows_X = X.iloc[-self.lag : None]
        else:
            self.last_rows_X = pd.concat([self.last_rows_X, X], axis="index").iloc[
                -self.lag : None
            ]

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if in_sample:
            return X.diff(self.lag)
        else:
            return (
                pd.concat([self.last_rows_X, X], axis="index")
                .diff(self.lag)
                .iloc[self.lag :]
            )

    def inverse_transform(self, X: pd.Series) -> pd.Series:
        return X.cumsum() + self.last_rows_X.iloc[0].squeeze()



from typing import Optional, Tuple

import pandas as pd


def check_types(
    X: Optional[pd.DataFrame], y: pd.Series
) -> Tuple[pd.DataFrame, pd.Series]:
    if X is None:
        X = pd.DataFrame(0, index=y.index, columns=["X_not_available"])
    else:
        assert isinstance(X, pd.DataFrame), "X must be a pandas DataFrame."
    assert isinstance(y, pd.Series), "y must be a pandas Series."
    return X, y



from typing import Optional, Union

import pandas as pd

from ..base import Transformation, fit_noop
from .base import Model


class WrapSKLearnClassifier(Model):
    """
    Wraps an SKLearn Classifier model.
    There's no need to use it directly, `fold` automatically wraps all sklearn classifiers into this class.
    """

    properties = Model.Properties(
        requires_X=True, model_type=Model.Properties.ModelType.classifier
    )

    def __init__(self, model) -> None:
        self.model = model
        self.name = model.__class__.__name__

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        if hasattr(self.model, "partial_fit"):
            self.model.partial_fit(X, y, sample_weights)
        else:
            self.model.fit(X, y, sample_weights)

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        if hasattr(self.model, "partial_fit"):
            self.model.partial_fit(X, y, sample_weights)
        # if we don't have partial_fit, we can't update the model (maybe throw an exception, and force user to wrap it into `DontUpdate`?)

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        probabilities = pd.DataFrame(
            data=self.model.predict_proba(X),
            index=X.index,
            columns=[
                f"probabilities_{self.name}_{item}" for item in self.model.classes_
            ],
        )
        predictions = pd.Series(
            data=self.model.predict(X).squeeze(),
            index=X.index,
            name=f"predictions_{self.name}",
        )
        return pd.concat([predictions, probabilities], axis="columns")

    predict_in_sample = predict


class WrapSKLearnRegressor(Model):
    """
    Wraps an SKLearn regressor model.
    There's no need to use it directly, `fold` automatically wraps all sklearn regressors into this class.
    """

    properties = Model.Properties(
        requires_X=True, model_type=Model.Properties.ModelType.regressor
    )

    def __init__(self, model) -> None:
        self.model = model
        self.name = model.__class__.__name__

    def fit(
        self, X: pd.DataFrame, y: pd.Series, sample_weights: Optional[pd.Series] = None
    ) -> None:
        if hasattr(self.model, "partial_fit"):
            self.model.partial_fit(X, y, sample_weights)
        else:
            self.model.fit(X, y, sample_weights)

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        if hasattr(self.model, "partial_fit"):
            self.model.partial_fit(X, y, sample_weights)
        # if we don't have partial_fit, we can't update the model (maybe throw an exception, and force user to wrap it into `DontUpdate`?)

    def predict(self, X: pd.DataFrame) -> Union[pd.Series, pd.DataFrame]:
        return pd.Series(
            data=self.model.predict(X).squeeze(),
            index=X.index,
            name=f"predictions_{self.name}",
        ).to_frame()

    predict_in_sample = predict


class WrapSKLearnPipeline(Model):
    """
    Wraps an scikit-learn Pipeline.
    It's usage is discouraged, as it's not possible to update an scikit-learn Pipeline with new data.
    Fold has all the primitives that scikit-learn Pipelines provide, just wrap your Transformations into an array.
    """

    properties = Transformation.Properties(requires_X=True)

    def __init__(self, pipeline) -> None:
        self.pipeline = pipeline
        self.name = pipeline.__class__.__name__

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        self.pipeline.fit(X, y)

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        return pd.Series(
            data=self.pipeline.predict(X).squeeze(),
            index=X.index,
            name=f"predictions_{self.name}",
        ).to_frame()

    predict_in_sample = predict
    update = fit_noop



from typing import Callable, List, Optional

import pandas as pd
import ray

from ...base import Composite, Transformations
from ...splitters import Fold
from ..types import Backend, Stage


def train_transformations(
    func: Callable,
    transformations: Transformations,
    X: pd.DataFrame,
    y: pd.Series,
    sample_weights: Optional[pd.Series],
    splits: List[Fold],
    never_update: bool,
    backend: Backend,
    silent: bool,
):
    func = ray.remote(func)
    X = ray.put(X)
    y = ray.put(y)
    return ray.get(
        [
            func.remote(
                X, y, sample_weights, transformations, split, never_update, backend
            )
            for split in splits
        ]
    )


def process_primary_child_transformations(
    func: Callable,
    list_of_child_transformations_with_index: List,
    composite: Composite,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
    stage: Stage,
    backend: Backend,
):
    func = ray.remote(func)
    X = ray.put(X)
    y = ray.put(y)
    return ray.get(
        [
            func.remote(
                composite,
                index,
                child_transformation,
                X,
                y,
                sample_weights,
                stage,
                backend,
            )
            for index, child_transformation in list_of_child_transformations_with_index
        ]
    )


def process_secondary_child_transformations(
    func: Callable,
    list_of_child_transformations_with_index: List,
    composite: Composite,
    X: pd.DataFrame,
    y: Optional[pd.Series],
    sample_weights: Optional[pd.Series],
    results_primary: List[pd.DataFrame],
    stage: Stage,
    backend: Backend,
):
    func = ray.remote(func)
    X = ray.put(X)
    y = ray.put(y)
    return ray.get(
        [
            func.remote(
                composite,
                index,
                child_transformation,
                X,
                y,
                sample_weights,
                results_primary,
                stage,
                backend,
            )
            for index, child_transformation in list_of_child_transformations_with_index
        ]
    )



import collections
from collections.abc import Iterable
from typing import List, Tuple, TypeVar, Union

from iteration_utilities import unique_everseen

T = TypeVar("T")


def wrap_in_list(input: Union[T, List[T]]) -> List[T]:
    return input if isinstance(input, List) else [input]


def transform_range_to_list(input: Union[range, List[T]]) -> List[T]:
    return list(input) if isinstance(input, range) else input


def wrap_in_double_list_if_needed(
    input: Union[T, List[T]]
) -> Union[List[List[T]], List[T]]:
    """
    If input is a single item, wrap it in a list.
    If input is a single list, wrap it in another list.
    If input is a list of lists, return it as is.
    """
    if not isinstance(input, List):
        return [input]
    if isinstance(input[0], List):
        return input
    else:
        return [input]


def flatten(input: Union[List[List], List]) -> Iterable:
    for x in input:
        if isinstance(x, list):
            yield from flatten(x)
        else:
            yield x


def keep_only_duplicates(input: List) -> List:
    return [item for item, count in collections.Counter(input).items() if count > 1]


def has_intersection(lhs: List, rhs: List) -> bool:
    return len(set(lhs).intersection(rhs)) > 0


def unique(input: List) -> List:
    return unique_everseen(input)


def swap_tuples(input: List[Tuple]) -> List[Tuple]:
    return [(b, a) for a, b in input]



from __future__ import annotations

from enum import Enum
from typing import Union


class Backend(Enum):
    """
    Parameters
    ----------
    no: string
        Uses sequential processing. This is the default.
    ray: string
        Uses `ray` as a backend. Call `ray.init()` before using this backend.

    """

    no = "no"
    ray = "ray"

    @staticmethod
    def from_str(value: Union[str, Backend]) -> Backend:
        if isinstance(value, Backend):
            return value
        for strategy in Backend:
            if strategy.value == value:
                return strategy
        else:
            raise ValueError(f"Unknown Backend: {value}")


class TrainMethod(Enum):
    """
    Parameters
    ----------
    parallel: string
        Parallel, independent training of pipelines for each fold.
    sequential: string
        Sequentially train/update pipelines, walking forward in time.

    """

    parallel = "parallel"
    sequential = "sequential"
    parallel_with_search = (  # Don't use it just yet, not yet fully documented
        "parallel_with_search"
    )

    @staticmethod
    def from_str(value: Union[str, TrainMethod]) -> TrainMethod:
        if isinstance(value, TrainMethod):
            return value
        for strategy in TrainMethod:
            if strategy.value == value:
                return strategy
        else:
            raise ValueError(f"Unknown TrainMethod: {value}")


class Stage(Enum):
    inital_fit = "inital_fit"
    update = "update"
    update_online_only = "update_online_only"
    infer = "infer"

    def is_fit_or_update(self) -> bool:
        return self in [Stage.inital_fit, Stage.update]



from inspect import getfullargspec
from typing import Callable, Optional

import pandas as pd

from ..base import InvertibleTransformation, Transformation, fit_noop


class Breakpoint(Transformation):
    """
    A transformation that stops execution at the specified point.
    """

    properties = Transformation.Properties(requires_X=False)

    def __init__(
        self,
        stop_at_fit: bool = True,
        stop_at_update: bool = True,
        stop_at_transform: bool = True,
    ) -> None:
        self.name = "Debug"
        self.stop_at_fit = stop_at_fit
        self.stop_at_update = stop_at_update
        self.stop_at_transform = stop_at_transform

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        if self.stop_at_fit:
            breakpoint()

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        if self.stop_at_update:
            breakpoint()

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if self.stop_at_transform:
            breakpoint()
        return X


class Identity(Transformation):
    properties = Transformation.Properties(requires_X=False)

    name = "Identity"

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        return X

    fit = fit_noop
    update = fit


class Test(InvertibleTransformation):
    properties = InvertibleTransformation.Properties(requires_X=False)
    __test__ = False
    no_of_calls_fit = 0
    no_of_calls_update = 0
    no_of_calls_transform_insample = 0
    no_of_calls_transform_outofsample = 0
    no_of_calls_inverse_transform = 0

    def __init__(
        self,
        fit_func: Callable,
        transform_func: Callable,
        update_func: Optional[Callable] = None,
        inverse_transform_func: Optional[Callable] = None,
    ) -> None:
        self.name = "Test"
        self.fit_func = fit_func
        self.transform_func = transform_func
        self.update_func = update_func
        self.inverse_transform_func = inverse_transform_func

    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        self.no_of_calls_fit += 1
        argspec = getfullargspec(self.fit_func)
        if len(argspec.args) == 1:
            self.fit_func(X)
        elif len(argspec.args) == 2:
            self.fit_func(X, y)
        elif len(argspec.args) == 3:
            self.fit_func(X, y, sample_weights)
        else:
            raise ValueError(
                "fit_func must accept between 1 and 3 arguments, "
                f"but {len(argspec.args)} were given."
            )

    def update(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        sample_weights: Optional[pd.Series] = None,
    ) -> None:
        self.no_of_calls_update += 1
        if self.update_func is None:
            return
        argspec = getfullargspec(self.update_func)
        if len(argspec.args) == 1:
            self.update_func(X)
        elif len(argspec.args) == 2:
            self.update_func(X, y)
        elif len(argspec.args) == 3:
            self.update_func(X, y, sample_weights)
        else:
            raise ValueError(
                "update_func must accept between 1 and 3 arguments, "
                f"but {len(argspec.args)} were given."
            )

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if in_sample:
            self.no_of_calls_transform_insample += 1
        else:
            self.no_of_calls_transform_outofsample += 1
        return_value = self.transform_func(X)
        if return_value is None:
            return X
        return return_value

    def inverse_transform(self, X: pd.Series) -> pd.Series:
        self.no_of_calls_inverse_transform += 1
        if self.inverse_transform_func is not None:
            return self.inverse_transform_func(X)
        return X



from __future__ import annotations

from typing import Callable, List, Optional

import pandas as pd

from ..base import Composite, Pipelines, Transformations
from .common import get_concatenated_names


class SelectBest(Composite):
    """
    Don't use this just yet. Coming later.
    """

    properties = Composite.Properties()

    def __init__(
        self,
        models: Pipelines,
        scorer: Callable,
        is_scorer_loss: bool = True,
        selected_model: Optional[Transformations] = None,
    ) -> None:
        self.models = models
        self.name = "SelectBest-" + get_concatenated_names(models)
        self.scorer = scorer
        self.is_scorer_loss = is_scorer_loss
        self.selected_model = selected_model

    def postprocess_result_primary(
        self, results: List[pd.DataFrame], y: Optional[pd.Series]
    ) -> pd.DataFrame:
        if self.selected_model is None:
            scores = [self.scorer(y, result) for result in results]
            selected_index = (
                scores.index(min(scores))
                if self.is_scorer_loss
                else scores.index(max(scores))
            )
            self.selected_model = [self.models[selected_index]]
            return results[selected_index]
        else:
            return results[0]

    def get_child_transformations_primary(self) -> Pipelines:
        if self.selected_model is None:
            return self.models
        else:
            return self.selected_model

    def clone(self, clone_child_transformations: Callable) -> SelectBest:
        return SelectBest(
            models=clone_child_transformations(self.models),
            scorer=self.scorer,
            is_scorer_loss=self.is_scorer_loss,
            selected_model=clone_child_transformations(self.selected_model)
            if self.selected_model is not None
            else None,
        )



from datetime import date
from enum import Enum
from typing import List, Union

import pandas as pd

from ..base import Transformation, fit_noop
from ..utils.list import swap_tuples, wrap_in_list


class LabelingMethod(Enum):
    """
    Parameters
    ----------
    holiday_binary: string
        * Workdays = 0
        * National Holidays = 1
    weekday_weekend_holiday: string
        * Workdays = 0
        * Weekends = 1
        * National Holidays == 2
    weekday_weekend_uniqueholiday: string
        * Workdays = 0
        * Weekends = 1
        * National Holidays == Unique int (>1)
    weekday_weekend_uniqueholiday_string: string
        * Workdays = 0
        * Weekends = 1
        * National Holidays == string
    """

    holiday_binary = "holiday_binary"
    weekday_weekend_holiday = "weekday_weekend_holiday"
    weekday_weekend_uniqueholiday = "weekday_weekend_uniqueholiday"
    weekday_weekend_uniqueholiday_string = "weekday_weekend_uniqueholiday_string"

    @staticmethod
    def from_str(value: Union[str, "LabelingMethod"]) -> "LabelingMethod":
        if isinstance(value, LabelingMethod):
            return value
        for strategy in LabelingMethod:
            if strategy.value == value:
                return strategy
        else:
            raise ValueError(f"Unknown HolidayType: {value}")


class AddHolidayFeatures(Transformation):
    """
    Adds holiday features for given region(s) as new column(s).
    It uses the pattern "holiday_{country_code}" for naming the columns.


    Parameters
    ----------

    country_codes: List[str]
        List of country codes  (eg.: `US`, `DE`) for which to add holiday features.
    labeling: LabelingMethod
        * holiday_binary: Workdays = 0 | National Holidays = 1
        * weekday_weekend_holiday: Workdays = 0 | Weekends = 1 | National Holidays == 2
        * weekday_weekend_uniqueholiday: Workdays = 0 | Weekends = 1 | National Holidays == Unique int (>1)
        * weekday_weekend_uniqueholiday_string: Workdays = 0 | Weekends = 1 | National Holidays == string

    """

    properties = Transformation.Properties(requires_X=False)

    def __init__(
        self,
        country_codes: Union[List[str], str],
        labeling: Union[str, LabelingMethod] = LabelingMethod.weekday_weekend_holiday,
    ) -> None:
        self.country_codes = [
            country_code.upper() for country_code in wrap_in_list(country_codes)
        ]
        self.name = f"AddHolidayFeatures-{self.country_codes}"
        self.type = LabelingMethod.from_str(labeling)
        from holidays import country_holidays, list_supported_countries

        all_supported_countries = list_supported_countries()

        assert all(
            [
                country_code in all_supported_countries
                for country_code in self.country_codes
            ]
        ), f"Country code not supported: {country_codes}"

        self.holiday_to_int_maps = [
            dict(
                swap_tuples(
                    enumerate(
                        sorted(
                            set(
                                country_holidays(
                                    country_code,
                                    years=list(range(1900, date.today().year)),
                                    language="en_US",
                                ).values()
                            )
                        ),
                        start=1,
                    )
                )
            )
            for country_code in self.country_codes
        ]

    def transform(self, X: pd.DataFrame, in_sample: bool) -> pd.DataFrame:
        if self.type == LabelingMethod.holiday_binary:
            holidays = _get_holidays(
                X.index, self.country_codes, self.holiday_to_int_maps, encode=True
            )
            holidays[holidays != 0] = 1

            return pd.concat([X, holidays], axis="columns")
        elif self.type == LabelingMethod.weekday_weekend_holiday:
            holidays = _get_holidays(
                X.index, self.country_codes, self.holiday_to_int_maps, encode=True
            )

            # All values that are (bigger than 0 or a string) are holidays, but we don't want to differentiate between them
            holidays[holidays != 0] = 2

            holidays[holidays == 0] = holidays[holidays == 0].add(
                _get_weekends(X.index), axis="index"
            )
            return pd.concat([X, holidays], axis="columns")
        elif (
            self.type == LabelingMethod.weekday_weekend_uniqueholiday
            or self.type == LabelingMethod.weekday_weekend_uniqueholiday_string
        ):
            holidays = _get_holidays(
                X.index,
                self.country_codes,
                self.holiday_to_int_maps,
                encode=self.type == LabelingMethod.weekday_weekend_uniqueholiday,
            )

            weekends = _get_weekends(X.index)
            holidays[holidays == 0] = holidays[holidays == 0].add(
                weekends, axis="index"
            )
            return pd.concat([X, holidays], axis="columns")
        else:
            raise ValueError(f"Unknown HolidayType: {self.type}")

    fit = fit_noop
    update = fit_noop


def _get_weekends(dates: pd.DatetimeIndex) -> pd.Series:
    return pd.Series((dates.weekday > 4).astype(int), index=dates)


def _get_holidays(
    dates: pd.DatetimeIndex,
    country_codes: List[str],
    holiday_to_int_maps: List[dict],
    encode: bool,
) -> pd.DataFrame:
    from holidays import country_holidays

    series = [
        pd.Series(
            country_holidays(
                country_code,
                years=dates.year.unique().to_list(),
                language="en_US",
            ),
            index=dates.date,
            name=f"holiday_{country_code}",
        ).set_axis(dates)
        for country_code in country_codes
    ]

    df = pd.concat(series, axis="columns").fillna(0)

    if encode:
        for country_code, holiday_to_int_map in zip(country_codes, holiday_to_int_maps):
            col = df[f"holiday_{country_code}"]
            if not col.dtype == "object":
                continue  # we don't use strings, there's nothing to encode
            df[f"holiday_{country_code}"] = (
                col.map(holiday_to_int_map).fillna(0).astype(int)
            )

    return df



from typing import List

import pandas as pd


def is_prediction(input: pd.DataFrame) -> bool:
    if len(input.columns) == 1:
        return (
            input.columns[0].startswith("predictions_")
            if isinstance(input.columns[0], str)
            else False
        )
    else:
        is_predictions_col_present = input.columns[0].startswith("predictions_")
        return is_predictions_col_present and all(
            [col.startswith("probabilities_") for col in input.columns[1:]]
        )


def all_have_probabilities(results: List[pd.DataFrame]) -> bool:
    """
    Check if all the DataFrames have probabilities columns,
    or if their values are all NaN, indicating an empty DataFrame
    being passed around, as SkipNA filtered out all data.
    """
    return all(
        [
            any([True for col in df.columns if col.startswith("probabilities_")])
            or len(df.columns) == 0
            or (len(df.columns) == 1 and df.isna().sum()[0] == len(df))
            for df in results
        ]
    )


def get_prediction_column(input: pd.DataFrame) -> pd.Series:
    return input[get_prediction_column_name(input)].squeeze()


def get_prediction_column_name(input: pd.DataFrame) -> str:
    candidates = [col for col in input.columns if col.startswith("predictions_")]
    if len(candidates) == 0:
        if len(input.columns) == 1:
            return input.columns[0]
        else:
            raise ValueError(f"Could not find a predictions column in {input.columns}.")
    return candidates[0]


def is_X_available(X: pd.DataFrame) -> bool:
    """
    Check if X is available, or the input is univariate, without exogenous variables.
    """
    return not (X.shape[1] == 1 and X.columns[0] == "X_not_available")


def is_columns_all(columns: List[str]):
    return columns[0] == "all"


def check_get_columns(columns: List[str], X: pd.DataFrame) -> List[str]:
    if is_columns_all(columns):
        return X.columns
    else:
        return columns
